import pandas as pd
import datetime
from mindsdb_sql.parser import ast
from pandas import DataFrame

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
)
from mindsdb.integrations.utilities.date_utils import utc_date_str_to_timestamp_ms, parse_utc_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class GoogleCalendarEventsTable(APITable):

    def select(self, query: ast.Select) -> DataFrame:
        """
        Gets all events from the calendar.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'timeMax' or arg1 == 'timeMin':
                date = parse_utc_date(arg2)
                if op == '=':
                    params[arg1] = date
                else:
                    raise NotImplementedError
            elif arg1 == 'timeZone':
                params[arg1] = arg2
            elif arg1 == 'maxAttendees':
                params[arg1] = arg2
            elif arg1 == 'q':
                params[arg1] = arg2
           

        # Get the order by from the query.
        if query.order_by is not None:
            if query.order_by[0].value == 'start_time':
                params['orderBy'] = 'startTime'
            elif query.order_by[0].value == 'updated':
                params['orderBy'] = 'updated'
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        # Get the events from the Google Calendar API.
        events = self.handler.call_application_api(method_name='get_events', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(events) == 0:
            events = pd.DataFrame([], columns=selected_columns)
        else:
            events.columns = self.get_columns()
            for col in set(events.columns).difference(set(selected_columns)):
                events = events.drop(col, axis=1)
        return events

    def insert(self, query: ast.Insert):
        """
        Inserts an event into the calendar.

        Args:
            query (ast.Insert): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Get the values from the query.
        values = query.values[0]
        # Get the event data from the values.
        event_data = {}
        timestamp_columns = {'start_time', 'end_time', 'created', 'updated'}
        regular_columns = {'summary', 'description', 'location', 'status', 'html_link',
                            'creator', 'organizer', 'reminders', 'timeZone', 'calendar_id', 'attendees'}

        # TODO: check why query.columns is None
        for col, val in zip(query.columns, values):
            if col.name in timestamp_columns:
                event_data[col.name] = utc_date_str_to_timestamp_ms(val)
            elif col.name in regular_columns:
                event_data[col.name] = val
            else:
                raise NotImplementedError
            
        st = datetime.datetime.utcfromtimestamp(event_data['start_time'] / 1000).isoformat() + 'Z'
        et = datetime.datetime.utcfromtimestamp(event_data['end_time'] / 1000).isoformat() + 'Z'

        event_data['start'] = {
            'dateTime': st,
            'timeZone': event_data['timeZone']
        }

        event_data['end'] = {
            'dateTime': et,
            'timeZone': event_data['timeZone']
        }

        event_data['attendees'] = event_data['attendees'].split(',')
        event_data['attendees'] = [{'email': attendee} for attendee in event_data['attendees']]

        # Insert the event into the Google Calendar API.
        self.handler.call_application_api(method_name='create_event', params=event_data)

    def update(self, query: ast.Update):
        """
        Updates an event or events in the calendar.

        Args:
            query (ast.Update): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Get the values from the query.
        values = query.values[0]
        # Get the event data from the values.
        event_data = {}
        for col, val in zip(query.update_columns, values):
            if col == 'start_time' or col == 'end_time' or col == 'created' or col == 'updated':
                event_data[col] = utc_date_str_to_timestamp_ms(val)
            elif col == 'summary' or col == 'description' or col == 'location' or col == 'status' or col == 'html_link' \
                    or col == 'creator' or col == 'organizer' or col == 'reminders' \
                    or col == 'timeZone' or col == 'calendar_id' or col == 'attendees':
                event_data[col] = val
            else:
                raise NotImplementedError

        event_data['start'] = {
            'dateTime': event_data['start_time'],
            'timeZone': event_data['timeZone']
        }

        event_data['end'] = {
            'dateTime': event_data['end_time'],
            'timeZone': event_data['timeZone']
        }

        event_data['attendees'] = event_data.get('attendees').split(',')
        event_data['attendees'] = [{'email': attendee} for attendee in event_data['attendees']]

        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if arg1 == 'event_id':
                if op == '=':
                    event_data['event_id'] = arg2
                elif op == '>':
                    event_data['start_id'] = arg2
                elif op == '<':
                    event_data['end_id'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        # Update the event in the Google Calendar API.
        self.handler.call_application_api(method_name='update_event', params=event_data)

    def delete(self, query: ast.Delete):
        """
        Deletes an event or events in the calendar.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'event_id':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['start_id'] = arg2
                elif op == '<':
                    params['end_id'] = arg2
                else:
                    raise NotImplementedError

        # Delete the events in the Google Calendar API.
        self.handler.call_application_api(method_name='delete_event', params=params)

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'etag',
            'id',
            'status',
            'htmlLink',
            'created',
            'updated',
            'summary',
            'creator',
            'organizer',
            'start',
            'end',
            'timeZone',
            'iCalUID',
            'sequence',
            'reminders',
            'eventType']



import pandas as pd
from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser import ast


class FrappeDocumentsTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the Frappe API and returns it as a pandas DataFrame.
        
        Returns dataframe representing the Frappe API results.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if arg1 == 'doctype':
                if op != '=':
                    raise NotImplementedError
                params['doctype'] = arg2
            elif arg1 == 'name':
                params['name'] = arg2
            else:
                filters.append([arg1, op, arg2])

        if not 'doctype' in params:
            raise ValueError('"doctype" parameter required')

        if query.limit:
            params['limit'] = query.limit.value
        if filters:
            params['filters'] = filters

        if 'name' in params:
            document_data = self.handler.call_frappe_api(
                method_name='get_document',
                params=params
            )
        else:
            document_data = self.handler.call_frappe_api(
                method_name='get_documents',
                params=params
            )

        # Only return the columns we need to.
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = document_data.columns
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(document_data) == 0:
            return pd.DataFrame([], columns=columns)

        # Remove columns not part of select.
        for col in set(document_data.columns).difference(set(columns)):
            document_data = document_data.drop(col, axis=1)

        return document_data

    def insert(self, query: ast.Insert) -> pd.DataFrame:
        columns = [col.name for col in query.columns]

        for row in query.values:
            params = dict(zip(columns, row))

            self.handler.call_frappe_api('create_document', params)

    def get_columns(self) -> List:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return ['doctype', 'data']



from enum import Enum


class ChatBotMessage:
    """
    Represents a message sent and received by chatbots.

    Attributes:
        type (ChatBotMessage.Type): Type of message
        text (str): Actual message content
        user (str): The user that sent the message
        destination (str): The user or channel that received the message
    """

    class Type(Enum):
        DIRECT = 1
        CHANNEL = 2

    def __init__(self, type: Type, text: str, user: str, destination: str):
        self.type = type
        self.text = text
        self.user = user
        self.destination = destination



from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.chatbot.chatbot_message import ChatBotMessage
from mindsdb.utilities.context import context as ctx


class RealtimeChatBotTask:
    """
    Manages the lifecycle and operation of a single chatbot.
    """

    _DEFAULT_MAX_ITERATIONS = 10
    _PROMPT_USER_COLUMN = 'input'

    def __init__(self, handler_factory, chat_engine, bot_record):

        # Need to set context first.
        self._bot_record = bot_record
        self._set_context()
        self._session = SessionController()

        self._chat_handler = handler_factory.create_realtime_chat_handler(chat_engine, self._on_message, bot_record.params)

        self._model_name = self._bot_record.model_name
        project_name = db.Project.query.get(self._bot_record.project_id).name
        self._project_datanode = self._session.datahub.get(project_name)

        # Get model info.
        model = self._session.model_controller.get_model(self._model_name, project_name=project_name)
        model_record = db.Predictor.query.get(model['id'])

        # Configure columns to use for responding.
        self._user_col = RealtimeChatBotTask._PROMPT_USER_COLUMN
        if 'using' in model_record.learn_args and 'user_column' in model_record.learn_args['using']:
            self._user_col = model_record.learn_args['using']['user_column']
        self._output_col = model_record.to_predict[0]

    def _set_context(self):
        ctx.set_default()
        ctx.company_id = self._bot_record.company_id
        if self._bot_record.user_class is not None:
            ctx.user_class = self._bot_record.user_class

    def run(self):
        """Begins listening and responding to messages"""
        self._chat_handler.connect()

    def stop(self):
        """Stops listening and responding to messages"""
        self._chat_handler.disconnect()

    def _on_message(self, message: ChatBotMessage):
        self._set_context()
        received_history = db.ChatBotsHistory(
            chat_bot_id=self._bot_record.id,
            type=message.type.name,
            text=message.text,
            user=message.user,
            destination='bot',
        )
        db.session.add(received_history)

        predict_params = {
            'max_iterations': RealtimeChatBotTask._DEFAULT_MAX_ITERATIONS,
            'verbose': True,
            'tools': []
        }
        if message.type == ChatBotMessage.Type.DIRECT:
            chatbot_response = self._project_datanode.predict(
                model_name=self._model_name,
                data={self._user_col: message.text},
                params=predict_params
            )

            response_text = chatbot_response.iloc[-1][self._output_col]
            response_message = ChatBotMessage(
                ChatBotMessage.Type.DIRECT,
                response_text,
                None,
                message.user
            )
            response = self._chat_handler.send_message(response_message)
            reply_history = db.ChatBotsHistory(
                chat_bot_id=self._bot_record.id,
                type=ChatBotMessage.Type.DIRECT.name,
                text=response_text,
                user='bot',
                destination=message.user,
            )
            if response.error:
                reply_history.error = response.error
            db.session.add(reply_history)
            db.session.commit()
            return

        raise NotImplementedError('Realtime chatbots only support direct messages currently')



import time
from mindsdb.utilities import log

from mindsdb.utilities.log import initialize_log

from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.chatbot.chatbot_thread import ChatBotThread
from mindsdb.interfaces.chatbot.realtime_chatbot_thread import RealtimeChatBotThread


class ChatBotMonitor:
    """Starts and stops chatbots based on their configuration in the database."""

    _MONITOR_INTERVAL_SECONDS = 1

    def __init__(self):
        self._active_bots = {}

    def start(self):
        """Starts monitoring chatbots periodically based on database configuration."""
        config = Config()
        db.init()
        initialize_log(config, 'jobs', wrap_print=True)
        self.config = config

        while True:
            try:
                self.check_bots()

            except (SystemExit, KeyboardInterrupt):
                self.stop_all_bots()
                raise
            except Exception as e:
                log.logger.error(e)

            db.session.rollback()  # disable cache
            time.sleep(ChatBotMonitor._MONITOR_INTERVAL_SECONDS)

    def stop_all_bots(self):
        """Stops all active bots."""
        active_bots = list(self._active_bots.keys())
        for bot_id in active_bots:
            self.stop_bot(bot_id)

    def check_bots(self):
        """Periodic check to see if new chatbots need to be started, or if existing chatbots need to be stopped."""
        allowed_bots = set()

        for bot in db.ChatBots.query.all():
            bot_id = bot.id
            allowed_bots.add(bot_id)

            if bot_id not in self._active_bots and bot.is_running:
                # Start new bot if configured to run.
                if bot.chat_engine is not None:
                    thread = RealtimeChatBotThread(bot)
                elif bot.database_id is not None:
                    thread = ChatBotThread(bot)
                else:
                    continue
                thread.start()
                self._active_bots[bot_id] = thread

        # Check old bots to stop
        active_bots = list(self._active_bots.keys())
        for bot_id in active_bots:
            bot = db.ChatBots.query.filter(
                db.ChatBots.id == bot_id
            ).first()
            is_running = bot is not None and bot.is_running
            if bot_id not in allowed_bots or not is_running:
                self.stop_bot(bot_id)

    def stop_bot(self, bot_id: int):
        """
        Stops a bot by id.

        Parameters:
            bot_id (int): ID of the bot to stop.
        """
        self._active_bots[bot_id].stop()
        del self._active_bots[bot_id]


def start(verbose=False):
    """Creates a new ChatBotMonitor based on MindsDB config and starts monitoring chatbots."""
    is_cloud = Config().get('cloud', False)
    if is_cloud is True:
        # Chatbots are disabled on cloud
        # TODO(tmichaeldb): Remove for public Chatbots release.
        return

    monitor = ChatBotMonitor()
    monitor.start()


if __name__ == '__main__':
    start()



import os
from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.handlers_wrapper.ml_grpc_wrapper import MLServiceServicer
from mindsdb.integrations.libs.handler_helpers import registry
from mindsdb.utilities.log import initialize_log, get_log


if __name__ == "__main__":
    config = Config()
    db.init()
    initialize_log(config=config)
    logger = get_log(logger_name="main")

    app = MLServiceServicer()
    port = int(os.environ.get("PORT", 5001))
    host = os.environ.get("HOST")

    # If there is a handler discovery service
    # start a thread to send service info
    # to the discovery service
    registry_url = os.environ.get("REGISTRY_URL")
    if registry_url:
        registry_url = f"{registry_url}/register"
        interval = os.environ.get("REGISTRY_INTERVAL", 50)
        service_type = os.environ.get("MINDSDB_SERVICE_TYPE", "lightwood")
        data = {"host": host, "port": port, "type": service_type}
        registry(registry_url, data, interval)

    logger.info("Running ML service: host=%s, port=%s", host, port)
    app.run(debug=True, host="0.0.0.0", port=port)



import threading

from mindsdb.utilities import log
from mindsdb.interfaces.chatbot.realtime_chatbot_task import RealtimeChatBotTask
from mindsdb.interfaces.chatbot.realtime_chat_handler_factory import RealtimeChatHandlerFactory


class RealtimeChatBotThread(threading.Thread):
    """A thread for a realtime chatbot to operate."""

    def __init__(self, bot_record):
        threading.Thread.__init__(self, daemon=True)
        self._bot_record = bot_record
        self._chatbot_task = None

    def run(self):
        """Starts running the chatbot"""
        self._chatbot_task = RealtimeChatBotTask(
            RealtimeChatHandlerFactory(),
            chat_engine=self._bot_record.chat_engine,
            bot_record=self._bot_record)

        try:
            self._chatbot_task.run()
        except Exception as e:
            log.logger.error(e)

    def stop(self):
        """Stops running the chatbot"""
        if self._chatbot_task:
            self._chatbot_task.stop()



from typing import Callable, Dict

from mindsdb.integrations.handlers.realtime_slack_chat_handler.realtime_slack_chat_handler import RealtimeSlackChatHandler
from mindsdb.integrations.libs.realtime_chat_handler import RealtimeChatHandler
from mindsdb.interfaces.chatbot.chatbot_message import ChatBotMessage


class RealtimeChatHandlerFactory:
    """Creates a RealtimeChatHandler based on configuration."""

    def create_realtime_chat_handler(self, chat_engine: str, on_message: Callable[[ChatBotMessage], None], params: Dict[str, str]) -> RealtimeChatHandler:
        """
        Creates a RealtimeChatHandler from the given chat engine and parameters.

        Parameters:
            chat_engine (str): The chat application to use
            on_message (Callable[[ChatBotMessage], None]): Callback for every message received
            params (Dict[str, str]): Parameters to pass to the handler

        Returns:
            handler (RealtimeChatHandler): Created chat handler
        """
        if chat_engine == 'slack':
            return RealtimeSlackChatHandler(
                on_message,
                params
            )
        raise NotImplementedError(f'Chat handler for engine {chat_engine} does not exist')



import json
import pandas as pd
import datetime as dt
from typing import Dict

from mindsdb.integrations.handlers.frappe_handler.frappe_tables import FrappeDocumentsTable
from mindsdb.integrations.handlers.frappe_handler.frappe_client import FrappeClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql


class FrappeHandler(APIHandler):
    """A class for handling connections and interactions with the Frappe API.

    Attributes:
        client (FrappeClient): The `FrappeClient` object for interacting with the Frappe API.
        is_connected (bool): Whether or not the API client is connected to Frappe.
        domain (str): Frappe domain to send API requests to.
        access_token (str): OAuth token to use for authentication.
    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.client = None
        self.is_connected = False

        args = kwargs.get('connection_data', {})
        if not 'access_token' in args:
            raise ValueError('"access_token" parameter required for authentication')
        if not 'domain' in args:
            raise ValueError('"domain" parameter required to connect to your Frappe instance')
        self.access_token = args['access_token']
        self.domain = args['domain']

        document_data = FrappeDocumentsTable(self)
        self._register_table('documents', document_data)
        self.connection_data = args

    def back_office_config(self):
        tools = {
            'register_sales_invoice': 'have to be used by assistant to register a sales invoice. Input is JSON object serialized as a string. Due date have to be passed in format: "yyyy-mm-dd".',
            'check_company_exists': 'useful to check the company is exist. Input is company',
            'check_expense_type': 'useful to check the expense_type is exist. Input is expense_type',
            'check_customer':  'useful to check the customer is exist. Input is customer',
            'check_item_code':  'have to be used to check the item code. Input is item_code',
        }
        return {
            'tools': tools,
        }

    def register_sales_invoice(self, data):
        """
          input is:
            {
              "due_date": "2023-05-31",
              "customer": "ksim",
              "items": [
                {
                  "name": "T-shirt--",
                  "description": "T-shirt",
                  "quantity": 1
                }
              ]
            }
        """
        invoice = json.loads(data)
        date = dt.datetime.strptime(invoice['due_date'], '%Y-%m-%d')
        if date <= dt.datetime.today():
            return 'Error: due_date have to be in the future'

        for item in invoice['items']:
            # rename column
            item['qty'] = item['quantity']
            del item['quantity']

            # add required fields
            item['uom'] = "Nos"
            item['conversion_factor'] = 1

            income_account = self.connection_data.get('income_account', "Sales Income - C8")
            item['income_account'] = income_account

        try:
            self.connect()
            self.client.post_document('Sales Invoice', invoice)
        except Exception as e:
            return f"Error: {e}"
        return f"Success"

    def check_item_code(self, item_code):
        self.connect()
        result = self.client.get_documents('Item', filters=[['item_code', '=', item_code]])
        if len(result) == 1:
            return True
        return "Item doesn't exist: please use different name"

    def check_company_exists(self, name):
        self.connect()
        result = self.client.get_documents('Company', filters=[['name', '=', name]])
        if len(result) == 1:
            return True
        return "Company doesn't exist: please use different name"

    def check_expense_type(self, name):
        self.connect()
        result = self.client.get_documents('Expense Claim Type', filters=[['name', '=', name]])
        if len(result) == 1:
            return True
        return "Expense Claim Type doesn't exist: please use different name"

    def check_customer(self, name):
        self.connect()
        result = self.client.get_documents('Customer', filters=[['name', '=', name]])
        if len(result) == 1:
            return True
        return "Customer doesn't exist"

    def connect(self) -> FrappeClient:
        """Creates a new  API client if needed and sets it as the client to use for requests.

        Returns newly created Frappe API client, or current client if already set.
        """
        if self.is_connected is True and self.client:
            return self.client

        if self.domain and self.access_token:
            self.client = FrappeClient(self.domain, self.access_token)

        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """Checks connection to Frappe API by sending a ping request.

        Returns StatusResponse indicating whether or not the handler is connected.
        """

        response = StatusResponse(False)

        try:
            client = self.connect()
            client.ping()
            response.success = True

        except Exception as e:
            log.logger.error(f'Error connecting to Frappe API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)

    def _document_to_dataframe_row(self, doctype, document: Dict) -> Dict:
        return {
            'doctype': doctype,
            'data': json.dumps(document)
        }

    def _get_document(self, params: Dict = None) -> pd.DataFrame:
        client = self.connect()
        doctype = params['doctype']
        document = client.get_document(doctype, params['name'])
        return pd.DataFrame.from_records([self._document_to_dataframe_row(doctype, document)])

    def _get_documents(self, params: Dict = None) -> pd.DataFrame:
        client = self.connect()
        doctype = params['doctype']
        limit = params.get('limit', None)
        filters = params.get('filters', None)
        fields = params.get('fields', None)
        documents = client.get_documents(doctype, limit=limit, fields=fields, filters=filters)
        return pd.DataFrame.from_records([self._document_to_dataframe_row(doctype, d) for d in documents])

    def _create_document(self, params: Dict = None) -> pd.DataFrame:
        client = self.connect()
        doctype = params['doctype']
        new_document = client.post_document(doctype, json.loads(params['data']))
        return pd.DataFrame.from_records([self._document_to_dataframe_row(doctype, new_document)])

    def call_frappe_api(self, method_name: str = None, params: Dict = None) -> pd.DataFrame:
        """Calls the Frappe API method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call (e.g. get_document)
            params (Dict): Params to pass to the API call
        """
        if method_name == 'get_documents':
            return self._get_documents(params)
        if method_name == 'get_document':
            return self._get_document(params)
        if method_name == 'create_document':
            return self._create_document(params)
        raise NotImplementedError('Method name {} not supported by Frappe API Handler'.format(method_name))



import logging

import pandas as pd

from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

from .urlcrawl_helpers import get_df_from_query_str, get_all_websites


class CrawlerTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:

        conditions = extract_comparison_conditions(query.where)
        urls = []
        for op, arg1, arg2 in conditions:

            if op == 'or':
                raise NotImplementedError(f'OR is not supported')

            if arg1 == 'url':
                url = arg2

                if op == '=':
                    urls = [str(url)]
                elif op == 'in':
                    if type(url) == str:
                        urls = [str(url)]
                    else:
                        urls = url
                else:
                    raise NotImplementedError(
                        f'url can be url = "someurl", you can also crawl multiple sites, as follows: url IN ("url1", "url2", ..)')

            else:
                pass

        if len(urls) == 0:
            raise NotImplementedError(
                f'You must specify what url you want to crawl, for example: SELECT * FROM crawl WHERE url IN ("someurl", ..)')

        limit = None

        if query.limit is not None:
            limit = query.limit.value
            if limit < 0:
                limit = None
        if limit is None:
            raise NotImplementedError(
                f'You must specify a LIMIT which defines how deep to crawl, a LIMIT -1 means that will crawl ALL websites and subwebsites (this can take a while)')

        result = get_all_websites(urls, limit, html=False)

        # filter targets
        result = project_dataframe(result, query.targets, self.get_columns())
        return result

    def get_columns(self):
        return [
            'url',
            'text_content',
            'error'
        ]


class WebHandler(APIHandler):
    """A class for handling crawling content from websites.

    Attributes:
        
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        self.api = None
        self.is_connected = True
        crawler = CrawlerTable(self)
        self._register_table('crawler', crawler)

    def check_connection(self) -> StatusResponse:

        response = StatusResponse(False)
        response.success = True

        return response

    def native_query(self, query_string: str = None):

        df = get_df_from_query_str(query_string)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)



import os
from mindsdb.utilities.log import get_log


logger = get_log(logger_name="main")


class DBClientFactory:
    def __init__(self):
        self.host = os.environ.get("MINDSDB_DB_SERVICE_HOST", None)
        self.port = os.environ.get("MINDSDB_DB_SERVICE_PORT", None)
        if self.host is not None and self.port is not None:
            try:
                from mindsdb.integrations.handlers_client.db_grpc_client import DBClientGRPC
                self.client_class = DBClientGRPC
            except (ImportError, ModuleNotFoundError):
                logger.info("to work with microservice mode please install 'pip install mindsdb[grpc]'")
                self.host = None
                self.port = None

    def __call__(self, handler_type: str, handler: type, **kwargs: dict):
        if self.host is None or self.port is None:
            logger.info(
                "%s.__call__: no post/port to DBService have provided. Handle all db request locally",
                self.__class__.__name__,
            )
            return handler(**kwargs)

        logger.info("%s.__call__: api to communicate with db services - gRPC, host - %s, port - %s",
                    self.__class__.__name__,
                    self.host,
                    self.port,
                    )

        return self.client_class(handler_type, **kwargs)


DBClient = DBClientFactory()



import sib_api_v3_sdk
import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable

from mindsdb_sql.parser import ast

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor


class EmailCampaignsTable(APITable):
    """The Sendinblue Email Campaigns Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Sendinblue "GET /emailCampaigns" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sendinblue Email Campaigns matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'email_campaigns',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        email_campaigns_df = pd.json_normalize(self.get_email_campaigns(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            email_campaigns_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        email_campaigns_df = select_statement_executor.execute_query()

        return email_campaigns_df

    def get_columns(self) -> List[str]:
        return pd.json_normalize(self.get_email_campaigns(limit=1)).columns.tolist()

    def get_email_campaigns(self, **kwargs):
        connection = self.handler.connect()
        email_campaigns_api_instance = sib_api_v3_sdk.EmailCampaignsApi(connection)
        email_campaigns = email_campaigns_api_instance.get_email_campaigns(**kwargs)
        return email_campaigns.campaigns


import smtplib
import imaplib
import email
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import pandas as pd


class EmailClient:
    def __init__(self, email, password, smtp_server='smtp.gmail.com', smtp_port=587, imap_server="imap.gmail.com"):
        self.email = email
        self.password = password
        self.smtp_server = smtplib.SMTP(smtp_server, smtp_port)
        self.imap_server = imaplib.IMAP4_SSL(imap_server)

    def send_email(self, to_addr, subject, body):
        msg = MIMEMultipart()
        msg['From'] = self.email
        msg['To'] = to_addr
        msg['Subject'] = subject
        msg.attach(MIMEText(body, 'plain'))

        self.smtp_server.starttls()
        self.smtp_server.login(self.email, self.password)
        self.smtp_server.send_message(msg)
        self.smtp_server.quit()

    def search_email(self, mailbox="INBOX", subject=None, to=None, from_=None, since_date=None, until_date=None,
                     since_emailid=None):

        self.imap_server.login(self.email, self.password)

        self.imap_server.select(mailbox)

        query_parts = []
        if subject is not None:
            query_parts.append(f'(SUBJECT "{subject}")')
        if to is not None:
            query_parts.append(f'(TO "{to}")')
        if from_ is not None:
            query_parts.append(f'(FROM "{from_}")')
        if since_date is not None:
            since_date_str = since_date.strftime("%d-%b-%Y")
            query_parts.append(f'(SINCE "{since_date_str}")')
        if until_date is not None:
            until_date_str = until_date.strftime("%d-%b-%Y")
            query_parts.append(f'(BEFORE "{until_date_str}")')
        if since_emailid is not None:
            query_parts.append(f'(UID {since_emailid}:*)')

        query = ' '.join(query_parts)

        ret = []

        resp, items = self.imap_server.uid('search', None, query)
        items = items[0].split()
        for emailid in items[::-1]:
            resp, data = self.imap_server.uid('fetch', emailid, "(BODY[HEADER.FIELDS (SUBJECT TO FROM DATE)])")
            try:
                raw_email = data[0][1].decode("utf-8")
            except UnicodeDecodeError:
                ValueError(f"Could not decode email with id {emailid}")
            email_message = email.message_from_string(raw_email)

            email_line = {}
            email_line['id'] = emailid
            email_line["to"] = email_message['To']
            email_line["from"] = email_message['From']
            email_line["subject"] = str(email_message['Subject'])
            email_line["created_at"] = email_message['Date']
            resp, email_data = self.imap_server.uid('fetch', emailid, '(BODY[TEXT])')
            email_line["body"] = email_data[0][1].decode('utf-8')

            ret.append(email_line)

        self.imap_server.logout()

        return pd.DataFrame(ret)







from typing import Dict, List

from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController

from mindsdb.utilities.context import context as ctx


class ChatBotController:
    '''Handles CRUD operations at the database level for Chatbots'''

    def __init__(self, project_controller: ProjectController = None):
        if project_controller is None:
            project_controller = ProjectController()
        self.project_controller = project_controller

    def _raise_if_on_cloud(self):
        # TODO(tmichaeldb): Remove for public Chatbots release.
        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            raise Exception('Chatbots are disabled on cloud')

    def get_chatbot(self, chatbot_name: str, project_name: str = 'mindsdb') -> db.ChatBots:
        '''
        Gets a chatbot by name.

        Parameters:
            chatbot_name (str): The name of the chatbot
            project_name (str): The name of the containing project

        Returns:
            bot (db.ChatBots): The database chatbot object
        '''
        self._raise_if_on_cloud()

        project = self.project_controller.get(name=project_name)
        bot = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.name == chatbot_name,
            db.ChatBots.project_id == project.id
        ).first()
        return bot

    def get_chatbots(self, project_name: str = 'mindsdb') -> List[db.ChatBots]:
        '''
        Gets all chatbots in a project.

        Parameters:
            project_name (str): The name of the containing project

        Returns:
            all_bots (List[db.ChatBots]): List of database chatbot object
        '''
        self._raise_if_on_cloud()

        project = self.project_controller.get(name=project_name)
        all_bots = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.project_id == project.id
        ).all()
        return all_bots

    def add_chatbot(
            self,
            name: str,
            project_name: str,
            model_name: str,
            database_id: int = None,
            chat_engine: str = None,
            is_running: bool = True,
            params: Dict[str, str] = {}) -> db.ChatBots:
        '''
        Adds a chatbot to the database.

        Parameters:
            name (str): The name of the new chatbot
            project_name (str): The containing project
            model_name (str): The name of the existing ML model the chatbot will use
            database_id (int): The ID of the existing database the chatbot will use
            chat_engine (str): The name of the chat integration the chatbot will use (e.g. slack)
            is_running (bool): Whether or not to start the chatbot right after creation
            params: (Dict[str, str]): Parameters to use when running the chatbot

        Returns:
            bot (db.ChatBots): The created chatbot
        '''
        self._raise_if_on_cloud()

        if project_name is None:
            project_name = 'mindsdb'
        project = self.project_controller.get(name=project_name)

        bot = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.name == name,
            db.ChatBots.project_id == project.id
        ).first()

        if bot is not None:
            raise Exception(f'Chat bot already exists: {name}')

        bot = db.ChatBots(
            company_id=ctx.company_id,
            name=name,
            project_id=project.id,
            model_name=model_name,
            database_id=database_id,
            chat_engine=chat_engine,
            is_running=is_running,
            params=params,
        )
        db.session.add(bot)
        db.session.commit()

        return bot

    def update_chatbot(
            self,
            chatbot_name: str,
            project_name: str = 'mindsdb',
            name: str = None,
            model_name: str = None,
            database_id: int = None,
            chat_engine: str = None,
            is_running: bool = None,
            params: Dict[str, str] = None):
        '''
        Updates a chatbot in the database, creating it if it doesn't already exist.

        Parameters:
            chatbot_name (str): The name of the new chatbot, or existing chatbot to update
            project_name (str): The containing project
            name (str): The updated name of the chatbot
            model_name (str): The name of the existing ML model the chatbot will use
            database_id (int): The ID of the existing database the chatbot will use
            chat_engine (str): The name of the chat integration the chatbot will use (e.g. slack)
            is_running (bool): Whether or not the chatbot will run after update/creation
            params: (Dict[str, str]): Parameters to use when running the chatbot

        Returns:
            bot (db.ChatBots): The created or updated chatbot
        '''
        self._raise_if_on_cloud()

        existing_chatbot = self.get_chatbot(chatbot_name, project_name=project_name)
        if existing_chatbot is None:
            return None

        if name is not None:
            existing_chatbot.name = name
        if model_name is not None:
            existing_chatbot.model_name = model_name
        if database_id is not None:
            existing_chatbot.database_id = database_id
        if chat_engine is not None:
            existing_chatbot.chat_engine = chat_engine
        if is_running is not None:
            existing_chatbot.is_running = is_running
        if params is not None:
            existing_chatbot.params = params
        db.session.commit()

        return existing_chatbot

    def delete_chatbot(self, chatbot_name: str, project_name: str = 'mindsdb'):
        '''
        Deletes a chatbot by name.

        Parameters:
            chatbot_name (str): The name of the chatbot to delete
            project_name (str): The name of the containing project
        '''

        self._raise_if_on_cloud()

        project = self.project_controller.get(name=project_name)

        bot = db.ChatBots.query.filter(
            db.ChatBots.company_id == ctx.company_id,
            db.ChatBots.name == chatbot_name,
            db.ChatBots.project_id == project.id
        ).first()

        if bot is None:
            raise Exception(f'Chat bot not found: {chatbot_name}')

        db.session.delete(bot)
        db.session.commit()



from flask_restx import Namespace, Resource
import mysql.connector


ns_conf = Namespace('nlp', description='The NLP API of mindsdb')


@ns_conf.route('/<query>')
@ns_conf.param('query', 'The natural language query')
@ns_conf.response(404, 'query failed')
class PredictorList(Resource):
    @ns_conf.doc('list_predictors')
    def get(self, query):
        '''Turn a natural language query into a sql query and deliver the response'''
        # For more info on ITG see https://github.com/mindsdb/ITG/tree/staging/itg
        try:
            from itg import itg
            # @TODO here and/or when this ns is started register all the user db schemas + mindsdb's db schema
            # itg.register((dataframe, 'db_name'))

            sql = itg(query).query
            print(f'The NLP API got the natural language query "{query}" and will execute the sql query: {sql}')

            con = mysql.connector.connect(
                host='localhost',
                port='47335',
                user='mindsdb',
                password='',
                database='mindsdb'
            )

            cur = con.cursor(dictionary=True, buffered=True)
            cur.execute(query)
            res = cur.fetchall()
            status = 200
        except Exception as e:
            res = {'error': str(e), 'sql_query': sql}
            status = 400
        con.commit()
        con.close()

        return res, status



from typing import Optional

import pandas as pd
import transformers
from huggingface_hub import HfApi

from mindsdb.utilities import log

from mindsdb.integrations.libs.base import BaseMLEngine


class HuggingFaceHandler(BaseMLEngine):
    name = 'huggingface'

    @staticmethod
    def create_validation(target, args=None, **kwargs):

        if 'using' in args:
            args = args['using']

        hf_api = HfApi()

        # check model is pytorch based
        metadata = hf_api.model_info(args['model_name'])
        if 'pytorch' not in metadata.tags:
            raise Exception('Currently only PyTorch models are supported (https://huggingface.co/models?library=pytorch&sort=downloads). To request another library, please contact us on our community slack (https://mindsdb.com/joincommunity).')

        # check model task
        supported_tasks = ['text-classification',
                           'zero-shot-classification',
                           'translation',
                           'summarization',
                           'text2text-generation',
                           'fill-mask']

        if metadata.pipeline_tag not in supported_tasks:
            raise Exception(f'Not supported task for model: {metadata.pipeline_tag}.\
             Should be one of {", ".join(supported_tasks)}')

        if 'task' not in args:
            args['task'] = metadata.pipeline_tag
        elif args['task'] != metadata.pipeline_tag:
            raise Exception(f'Task mismatch for model: {args["task"]}!={metadata.pipeline_tag}')

        input_keys = list(args.keys())

        # task, model_name, input_column is essential
        for key in ['task', 'model_name', 'input_column']:
            if key not in args:
                raise Exception(f'Parameter "{key}" is required')
            input_keys.remove(key)

        # check tasks input

        if args['task'] == 'zero-shot-classification':
            key = 'candidate_labels'
            if key not in args:
                raise Exception('"candidate_labels" is required for zero-shot-classification')
            input_keys.remove(key)

        if args['task'] == 'translation':
            keys = ['lang_input', 'lang_output']
            for key in keys:
                if key not in args:
                    raise Exception(f'{key} is required for translation')
                input_keys.remove(key)

        if args['task'] == 'summarization':
            keys = ['min_output_length', 'max_output_length']
            for key in keys:
                if key not in args:
                    raise Exception(f'{key} is required for translation')
                input_keys.remove(key)

        # optional keys
        for key in ['labels', 'max_length', 'truncation_policy']:
            if key in input_keys:
                input_keys.remove(key)

        if len(input_keys) > 0:
            raise Exception(f'Not expected parameters: {", ".join(input_keys)}')

    def create(self, target, args=None, **kwargs):
        # TODO change BaseMLEngine api?
        if 'using' in args:
            args = args['using']

        args['target'] = target

        model_name = args['model_name']
        hf_model_storage_path = self.engine_storage.folder_get(model_name)  # real

        if args['task'] == 'translation':
            args['task_proper'] = f"translation_{args['lang_input']}_to_{args['lang_output']}"
        else:
            args['task_proper'] = args['task']

        log.logger.debug(f"Checking file system for {model_name}...")

        ####
        # Check if pipeline has already been downloaded
        try:
            pipeline = transformers.pipeline(task=args['task_proper'], model=hf_model_storage_path,
                                             tokenizer=hf_model_storage_path)
            log.logger.debug(f'Model already downloaded!')
        ####
        # Otherwise download it
        except OSError:
            try:
                log.logger.debug(f"Downloading {model_name}...")
                pipeline = transformers.pipeline(task=args['task_proper'], model=model_name)

                pipeline.save_pretrained(hf_model_storage_path)

                log.logger.debug(f"Saved to {hf_model_storage_path}")
            except Exception:
                raise Exception("Error while downloading and setting up the model. Please try a different model. We're working on expanding the list of supported models, so we would appreciate it if you let us know about this in our community slack (https://mindsdb.com/joincommunity).")  # noqa
        ####

        if 'max_length' in args:
            pass
        elif 'max_position_embeddings' in pipeline.model.config.to_dict().keys():
            args['max_length'] = pipeline.model.config.max_position_embeddings
        elif 'max_length' in pipeline.model.config.to_dict().keys():
            args['max_length'] = pipeline.model.config.max_length
        else:
            log.logger.debug('No max_length found!')

        labels_default = pipeline.model.config.id2label
        labels_map = {}
        if 'labels' in args:
            for num in labels_default.keys():
                labels_map[labels_default[num]] = args['labels'][num]
            args['labels_map'] = labels_map
        else:
            for num in labels_default.keys():
                labels_map[labels_default[num]] = labels_default[num]
            args['labels_map'] = labels_map

        ###### store and persist in model folder
        self.model_storage.json_set('args', args)

        ###### persist changes to handler folder
        self.engine_storage.folder_sync(model_name)

    def predict_text_classification(self, pipeline, item, args):
        top_k = args.get('top_k', 1000)

        result = pipeline([item], top_k=top_k, truncation=True, max_length=args['max_length'])[0]

        final = {}
        explain = {}
        if type(result) == dict:
            result = [result]
        final[args['target']] = args['labels_map'][result[0]['label']]
        for elem in result:
            if args['labels_map']:
                explain[args['labels_map'][elem['label']]] = elem['score']
            else:
                explain[elem['label']] = elem['score']
        final[f"{args['target']}_explain"] = explain
        return final

    def predict_zero_shot(self, pipeline, item, args):
        top_k = args.get('top_k', 1000)

        result = pipeline([item], candidate_labels=args['candidate_labels'],
                                     truncation=True, top_k=top_k, max_length=args['max_length'])[0]

        final = {}
        final[args['target']] = result['labels'][0]

        explain = dict(zip(result['labels'], result['scores']))
        final[f"{args['target']}_explain"] = explain

        return final

    def predict_translation(self, pipeline, item, args):
        result = pipeline([item], max_length=args['max_length'])[0]

        final = {}
        final[args['target']] = result['translation_text']

        return final

    def predict_summarization(self, pipeline, item, args):
        result = pipeline([item], min_length=args['min_output_length'], max_length=args['max_output_length'])[0]

        final = {}
        final[args['target']] = result['summary_text']

        return final

    def predict_text2text(self, pipeline, item, args):
        result = pipeline([item], max_length=args['max_length'])[0]

        final = {}
        final[args['target']] = result['generated_text']

        return final

    def predict_fill_mask(self, pipeline, item, args):
        result = pipeline([item])[0]

        final = {}
        final[args['target']] = result[0]['sequence']
        explain = {elem['sequence']: elem['score'] for elem in result}
        final[f"{args['target']}_explain"] = explain

        return final

    def predict(self, df, args=None):

        fnc_list = {
            'text-classification': self.predict_text_classification,
            'zero-shot-classification': self.predict_zero_shot,
            'translation': self.predict_translation,
            'summarization': self.predict_summarization,
            'fill-mask': self.predict_fill_mask
        }

        ###### get stuff from model folder
        args = self.model_storage.json_get('args')

        task = args['task']

        if task not in fnc_list:
            raise RuntimeError(f'Unknown task: {task}')

        fnc = fnc_list[task]

        hf_model_storage_path = self.engine_storage.folder_get(args['model_name'], update=False)

        pipeline = transformers.pipeline(task=args['task_proper'], model=hf_model_storage_path,
                                         tokenizer=hf_model_storage_path)

        input_column = args['input_column']
        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')
        input_list = df[input_column]

        max_tokens = pipeline.tokenizer.model_max_length

        results = []
        for item in input_list:
            if max_tokens is not None:
                tokens = pipeline.tokenizer.encode(item)
                if len(tokens) > max_tokens:
                    truncation_policy = args.get('truncation_policy', 'strict')
                    if truncation_policy == 'strict':
                        results.append({'error': f'Tokens count exceed model limit: {len(tokens)} > {max_tokens}'})
                        continue
                    elif truncation_policy == 'left':
                        tokens = tokens[-max_tokens + 1: -1]  # cut 2 empty tokens from left and right
                    else:
                        tokens = tokens[1: max_tokens - 1]  # cut 2 empty tokens from left and right

                    item = pipeline.tokenizer.decode(tokens)

            item = str(item)
            try:
                result = fnc(pipeline, item, args)
            except Exception as e:
                msg = str(e).strip()
                if msg == '':
                    msg = e.__class__.__name__
                result = {'error': msg}
            results.append(result)

        pred_df = pd.DataFrame(results)

        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        args = self.model_storage.json_get('args')

        if attribute == 'args':
            return pd.DataFrame(args.items(), columns=['key', 'value'])
        elif attribute == 'metadata':
            hf_api = HfApi()
            metadata = hf_api.model_info(args['model_name'])
            data = metadata.__dict__
            return pd.DataFrame(list(data.items()), columns=['key', 'value'])
        else:
            tables = ['args', 'metadata']
            return pd.DataFrame(tables, columns=['tables'])



"""jobs_user_class

Revision ID: ef04cdbe51ed
Revises: 3154382dab17
Create Date: 2023-02-27 11:32:35.298375

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa


# revision identifiers, used by Alembic.
revision = 'ef04cdbe51ed'
down_revision = '3154382dab17'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('user_class', sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('user_class')

    # ### end Alembic commands ###



import time
from datetime import datetime, timezone
from functools import wraps

import mindsdb.utilities.hooks as hooks
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx


def _get_current_node(profiling: dict) -> dict:
    """ return the node that the pointer points to

        Args:
            profiling (dict): whole profiling data

        Returns:
            dict: current node
    """
    current_node = profiling['tree']
    for child_index in profiling['pointer']:
        current_node = current_node['children'][child_index]
    return current_node


def start_node(tag: str):
    """ Add a new node to profiling

        Args:
            tag (str): name of new node
    """
    profiling = ctx.profiling
    new_node = {
        'start_at': time.perf_counter(),
        'start_at_thread': time.thread_time(),
        'start_at_process': time.process_time(),
        'stop_at': None,
        'name': tag,
        'children': []
    }
    if profiling['pointer'] is None:
        if profiling['level'] != 1:
            # profiling was activated not in the root of nodes tree
            return
        profiling['pointer'] = []
        profiling['tree'] = new_node
        profiling['tree']['time_start_at'] = datetime.now(timezone.utc)
    else:
        current_node = _get_current_node(profiling)
        profiling['pointer'].append(len(current_node['children']))
        current_node['children'].append(new_node)


def stop_current_node():
    """ Mark current node as completed and move pointer up
    """
    profiling = ctx.profiling
    if profiling['pointer'] is None:
        # profiling was activated not in the root of nodes tree
        return
    current_node = _get_current_node(profiling)
    current_node['stop_at'] = time.perf_counter()
    current_node['stop_at_thread'] = time.thread_time()
    current_node['stop_at_process'] = time.process_time()
    if len(profiling['pointer']) > 0:
        profiling['pointer'] = profiling['pointer'][:-1]
    else:
        if ctx.profiling['enabled'] is True:
            _send_profiling_results()
        profiling['pointer'] = None


def set_meta(**kwargs):
    """ Add any additional info to profiling data

        Args:
            **kwargs (dict): metadata to add
    """
    if profiling_enabled() is True:
        ctx.profiling.update(kwargs)


def _send_profiling_results():
    """ Send profiling results to storage
    """
    ctx.profiling['company_id'] = ctx.company_id
    ctx.profiling['hostname'] = Config().get('aws_meta_data', {}).get('public-hostname', '?')
    ctx.profiling['instance_id'] = Config().get('aws_meta_data', {}).get('instance-id', '?')
    hooks.send_profiling_results(ctx.profiling)


def enable():
    ctx.profiling['enabled'] = True


def disable():
    ctx.profiling['enabled'] = False


def profiling_enabled():
    return ctx.profiling['enabled'] is True


def start(tag):
    """ add new node to profiling data
    """
    ctx.profiling['level'] += 1
    if profiling_enabled() is True:
        start_node(tag)


def stop():
    """ finalize current node and move pointer up
    """
    ctx.profiling['level'] -= 1
    if profiling_enabled() is True:
        stop_current_node()


class Context:
    def __init__(self, tag):
        self.tag = tag

    def __enter__(self):
        start(self.tag)

    def __exit__(self, exc_type, exc_value, traceback):
        stop()


def profile(tag: str = None):
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if profiling_enabled() is True:
                with Context(tag or f'{function.__name__}|{function.__module__}'):
                    result = function(*args, **kwargs)
            else:
                result = function(*args, **kwargs)
            return result
        return wrapper
    return decorator



import unittest
from mindsdb.integrations.handlers.informix_handler.informix_handler import InformixHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class InformixHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "server": "server",
                "host": "127.0.0.1",
                "port": 9093,
                "user": "informix",
                "password": "in4mix",
                "database": "demo",
                "schema_name": "love",
                "loging_enabled": False
            }
        }
        cls.handler = InformixHandler('test_informix_handler', cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_drop_table(self):
        res = self.handler.query("DROP TABLE IF EXISTS LOVE;")
        assert res.type is  not RESPONSE_TYPE.ERROR

    def test_2_create_table(self):
        res = self.handler.query("CREATE TABLE IF NOT EXISTS LOVE (LOVER varchar(20));")
        assert res.type is not  RESPONSE_TYPE.ERROR 
    
    def test_3_insert(self):
        res = self.handler.query("INSERT INTO LOVE VALUES('Hari');")
        assert res.type is not RESPONSE_TYPE.ERROR
    
    def test_4_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is RESPONSE_TYPE.TABLE
 
    def test_5_select_query(self):
        query = "SELECT * FROM LOVE;"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_check_connection(self):
        self.handler.check_connection()

        
if __name__ == '__main__':
    unittest.main()


import unittest
from mindsdb.integrations.handlers.google_calendar_handler.google_calendar_handler import GoogleCalendarHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class GoogleCalendarHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "credentials": "C:\\Users\\panagiotis\\Desktop\\GitHub\\mindsdb\\mindsdb\\integrations\\handlers"
                               "\\google_calendar_handler\\credentials.json",
            }
        }
        cls.handler = GoogleCalendarHandler('test_google_calendar_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_query(self):
        query = "SELECT summary FROM calendar.events WHERE id = 1"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_get_columns(self):
        columns = self.handler.get_columns('id')
        assert columns.type is not RESPONSE_TYPE.ERROR

    def test_4_insert(self):
        res = self.handler.native_query(
            "INSERT INTO calendar.events VALUES (100, '2023-04-21 00:00:00', '2023-05-01 00:00:00',"
            "'summary', 'description','location', 'status', 'html_link', "
            "'creator', 'organizer', 'reminders', "
            "'timeZone', 'calendar_id', 'attendees'")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_5_update(self):
        res = self.handler.native_query("UPDATE calendar.events SET summary = 'ONE HUNDRED' WHERE id = 100")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_6_delete(self):
        res = self.handler.native_query("DELETE FROM calendar.events WHERE id = 100")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_7_delete_multiple(self):
        res = self.handler.native_query("DELETE FROM calendar.events WHERE id > 1 AND id < 20")
        assert res.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import psycopg2
from pandas import DataFrame
from typing import Optional
from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import PostgresHandler



class YugabyteHandler(PostgresHandler):
    """
    This handler handles connection and execution of the YugabyteSQL statements.
    """
    name = 'yugabyte'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the YugabyteDB server/database.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the YugabyteDB server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the YugabyteDB server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect YugabyteDB server'
    }, 
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name  to connect YugabyteDB server'
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': '(OPTIONAL) comma seperated value of schema to be considered while querying'
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': ''' (OPTIONAL) Specifies the SSL mode for the connection, determining whether to use SSL encryption and the level of verification required 
        (e.g., "**disable**", "**allow**", "**prefer**", "**require**", "**verify-ca**", "**verify-full**")'''
    },


)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=5433,
    password='',
    user='admin',
    database='yugabyte'
    

)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import math
import struct

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import ONE_BYTE_ENC, TWO_BYTE_ENC, THREE_BYTE_ENC, NULL_VALUE, DEFAULT_CAPABILITIES
from mindsdb.api.mysql.mysql_proxy.utilities import logger


class Datum():
    def __init__(self, type, value=None):
        # TODO other types: float, timestamp
        self.type = type
        self.value = b''
        self.var_type = self.type.split('<')[0]
        self.var_len = self.type.split('<')[1].replace('>', '')

        if value is not None:
            self.set(value)

    def set(self, value):
        self.value = value

    def setFromBuff(self, buff):
        start = 0

        if self.var_len == 'lenenc':
            start = 1
            ln_enc = buff[0]
            if int(ln_enc) <= ONE_BYTE_ENC[0]:
                start = 0
                end = 1
            elif int(ln_enc) == TWO_BYTE_ENC[0]:
                end = 3
            elif int(ln_enc) == THREE_BYTE_ENC[0]:
                end = 4
            elif ln_enc:
                end = 9

            num_str = buff[start:end]
            if end > 9:
                logger.error('Cant decode integer greater than 8 bytes')
                return buff[end - 1:]

            for j in range(8 - (end - start)):
                num_str += b'\0'

            if self.var_type == 'int':
                self.value = struct.unpack('i', num_str)
                return buff[end:]

            if self.var_type in ['byte', 'string']:
                length = struct.unpack('Q', num_str)[0]
                self.value = buff[end:(length + end)]
                return buff[(length + end):]

        if self.var_len == 'EOF':
            length = len(buff)
            self.var_len = str(length)
            self.value = buff
            return ''
        else:
            length = self.var_len

        if self.type == 'string<NUL>':
            for j, x in enumerate(buff):
                if int(x) == 0:
                    length = j + 1
                    break

        length = int(length)
        if self.var_type in ['byte', 'string']:
            end = length
            self.value = buff[:end]
        else:  # if its an integer
            end = length
            num_str = buff[:end]
            if end > 8:
                logger.error('cant decode integer greater than 8 bytes')
                return buff[end:]
            for j in range(8 - end):
                num_str += b'\0'
            self.value = struct.unpack('Q', num_str)[0]
        if str(self.var_len) == 'NUL':
            self.value = self.value[:-1]
        return buff[end:]

    def lenencInt(self, value):
        byte_count = int(math.ceil(math.log((value + 1), 2) / 8))
        if byte_count == 0:
            return b'\0'
        if value < NULL_VALUE[0]:
            return struct.pack('i', value)[:1]
        if value >= NULL_VALUE[0] and byte_count <= 2:
            return TWO_BYTE_ENC + struct.pack('i', value)[:2]
        if byte_count <= 3:
            return THREE_BYTE_ENC + struct.pack('i', value)[:3]
        if byte_count <= 8:
            return THREE_BYTE_ENC + struct.pack('Q', value)[:8]

    def toStringPacket(self):
        if self.type == 'string<packet>':
            return self.value.get_packet_string()

        if self.type in ['string<EOF>', 'byte<EOF>']:
            length = int(len(self.value))
            self.var_len = length
            if length == 0:
                return b''
            else:
                return struct.pack('{len}s'.format(len=self.var_len), bytes(self.value, 'utf-8'))[:length]

        if self.type == 'string<NUL>':
            return bytes(self.value, 'utf-8') + struct.pack('b', 0)

        if self.var_len.isdigit():
            length = int(self.var_len)

            if self.var_type == 'int':
                # little endian format
                return struct.pack('Q', self.value)[:length]
            if self.var_type == 'string':
                return struct.pack(self.var_len + 's', bytes(self.value, 'utf-8'))[:length]
            if self.var_type == 'byte':
                return struct.pack(self.var_len + 's', self.value)[:length]

        elif self.var_len == 'lenenc':
            if self.value is None:
                return NULL_VALUE
            if self.var_type == 'int':
                return self.lenencInt(self.value)

            if self.var_type in ['byte', 'string']:
                value = self.value.decode() if isinstance(self.value, bytes) else self.value
                if isinstance(value, str) is False and value is not None:
                    value = str(value)

                val_len = len(value.encode('utf8'))

                byte_count = int(math.ceil(math.log((val_len + 1), 2) / 8))
                if val_len < NULL_VALUE[0]:
                    return self.lenencInt(val_len) + bytes(value, 'utf-8')
                if val_len >= NULL_VALUE[0] and byte_count <= 2:
                    return TWO_BYTE_ENC + struct.pack('i', val_len)[:2] + bytes(value, 'utf-8')
                if byte_count <= 3:
                    return THREE_BYTE_ENC + struct.pack('i', val_len)[:3] + bytes(value, 'utf-8')
                if byte_count <= 8:
                    return THREE_BYTE_ENC + struct.pack('Q', val_len)[:8] + bytes(value, 'utf-8')


def test():
    import pprint

    u = Datum('int<8>', DEFAULT_CAPABILITIES >> 16)
    pprint.pprint(u.toStringPacket())


# only run the test if this file is called from debugger
if __name__ == "__main__":
    test()



import re
from functools import lru_cache
from typing import List, Union

import pandas as pd
import torch
from langchain.document_loaders import DataFrameLoader
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter

from mindsdb.utilities import log


def extract_collection_name(sql_query):
    # Regular expression pattern to match the collection name from the FROM clause
    pattern = r"FROM\s+\w+\.(\w+)"

    # Find the table name using regular expression
    match = re.search(pattern, sql_query, re.IGNORECASE)

    # If match found then extract the table name

    if match:
        table_name = match.group(1)
        where_condition = None

        # Regular expression pattern to match the where condition from the WHERE clause
        pattern = r"WHERE\s+(.*)"
        # Find the where condition using regular expression
        match = re.search(pattern, sql_query, re.IGNORECASE)
        if match:
            where_condition = match.group(1)

        return table_name, where_condition
    else:
        return None, None


def get_metadata_filter(metadata_filter: str):
    """convert metadata filter string to dict"""
    dict_from_string = {}

    for item in metadata_filter.split(","):
        if ":" in item:
            key, value = item.split(":")
            dict_from_string[key] = value
        else:
            key = item
            value = True
            dict_from_string[key] = value

    return dict_from_string


# Todo move all below classes methods to dataprep ML
class DfLoader(DataFrameLoader):

    """
    override the load method of langchain.document_loaders.DataFrameLoaders to ignore rows with 'None' values
    """

    def __init__(self, data_frame: pd.DataFrame, page_content_column: str):
        super().__init__(data_frame=data_frame, page_content_column=page_content_column)
        self._data_frame = data_frame
        self._page_content_column = page_content_column

    def load(self) -> List[Document]:
        """Loads the dataframe as a list of documents"""
        documents = []
        for n_row, frame in self._data_frame[self._page_content_column].iteritems():
            if pd.notnull(frame):
                # ignore rows with None values
                column_name = self._page_content_column

                document_contents = frame

                documents.append(
                    Document(
                        page_content=document_contents,
                        metadata={
                            "source": "dataframe",
                            "row": n_row,
                            "column": column_name,
                        },
                    )
                )
        return documents


def df_to_documents(
    df: pd.DataFrame, page_content_columns: Union[List[str], str]
) -> List[Document]:
    """Converts a given dataframe to a list of documents"""
    documents = []

    if isinstance(page_content_columns, str):
        page_content_columns = [page_content_columns]

    for _, page_content_column in enumerate(page_content_columns):
        if page_content_column not in df.columns.tolist():
            raise ValueError(
                f"page_content_column {page_content_column} not in dataframe columns"
            )

        loader = DfLoader(data_frame=df, page_content_column=page_content_column)
        documents.extend(loader.load())

    return documents


def split_documents(df, columns):
    # Load documents and split in chunks
    log.logger.info(f"Loading documents from input data")

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=500, chunk_overlap=50)
    documents = df_to_documents(df=df, page_content_columns=columns)
    texts = text_splitter.split_documents(documents)
    log.logger.info(f"Loaded {len(documents)} documents from input data")
    log.logger.info(f"Split into {len(texts)} chunks of text (max. 500 tokens each)")

    return texts


@lru_cache()
def load_embeddings_model(embeddings_model_name):
    try:
        model_kwargs = {"device": "gpu" if torch.cuda.is_available() else "cpu"}
        embedding_model = HuggingFaceEmbeddings(
            model_name=embeddings_model_name, model_kwargs=model_kwargs
        )
    except ValueError:
        raise ValueError(
            f"The {embeddings_model_name}  is not supported, please select a valid option from Hugging Face Hub!"
        )
    return embedding_model


def documents_to_df(documents: list):
    """Converts a list of documents to a dataframe"""
    df = pd.DataFrame()
    df["page_content"] = [document.page_content for document in documents]
    df["metadata"] = [document.metadata for document in documents]

    return df



import time

from flask import request, session
from flask_restx import Resource
from flask_restx import fields

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.namespaces.configs.default import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import get_log


log = get_log('http')


def check_auth() -> bool:
    ''' checking whether current user is authenticated

        Returns:
            bool: True if user authentication is approved
    '''
    config = Config()
    if config['auth']['http_auth_enabled'] is False:
        return True

    if config['auth'].get('provider') == 'cloud':
        if isinstance(session.get('username'), str) is False:
            return False

        if config['auth']['oauth']['tokens']['expires_at'] < time.time():
            return False

        return True

    return session.get('username') == config['auth']['username']


@ns_conf.route('/login', methods=['POST'])
class LoginRoute(Resource):
    @ns_conf.doc(
        responses={
            200: 'Success',
            400: 'Error in username or password',
            401: 'Invalid username or password'
        },
        body=ns_conf.model('request_login', {
            'username': fields.String(description='Username'),
            'password': fields.String(description='Password')
        })
    )
    def post(self):
        ''' Check user's credentials and creates a session
        '''
        username = request.json.get('username')
        password = request.json.get('password')
        if (
            isinstance(username, str) is False or len(username) == 0
            or isinstance(password, str) is False or len(password) == 0
        ):
            return http_error(
                400, 'Error in username or password',
                'Username and password should be string'
            )

        config = Config()
        inline_username = config['auth']['username']
        inline_password = config['auth']['password']

        if (
            username != inline_username
            or password != inline_password
        ):
            return http_error(
                401, 'Forbidden',
                'Invalid username or password'
            )

        session.clear()
        session['username'] = username
        session.permanent = True

        return '', 200


@ns_conf.route('/logout', methods=['POST'])
class LogoutRoute(Resource):
    @ns_conf.doc(
        responses={
            200: 'Success'
        }
    )
    def post(self):
        session.clear()
        return '', 200


@ns_conf.route('/status')
class StatusRoute(Resource):
    @ns_conf.doc(
        responses={
            200: 'Success'
        },
        model=ns_conf.model('response_status', {
            'environment': fields.String(description='The name of current environment: cloud, local or other'),
            'mindsdb_version': fields.String(description='Current version of mindsdb'),
            'auth': fields.Nested(
                ns_conf.model('response_status_auth', {
                    'confirmed': fields.Boolean(description='is current user autentificated'),
                    'required': fields.Boolean(description='is autentificated required'),
                    'provider': fields.Boolean(description='current autentification provider: local of 3d-party')
                })
            )
        })
    )
    def get(self):
        ''' returns auth and environment data
        '''
        environment = 'local'
        config = Config()

        environment = config.get('environment')
        if environment is None:
            if config.get('cloud', False):
                environment = 'cloud'
            elif config.get('aws_marketplace', False):
                environment = 'aws_marketplace'
            else:
                environment = 'local'

        auth_provider = 'disabled'
        if config['auth']['http_auth_enabled'] is True:
            if config['auth'].get('provider') is not None:
                auth_provider = config['auth'].get('provider')
            else:
                auth_provider = 'local'

        resp = {
            'mindsdb_version': mindsdb_version,
            'environment': environment,
            'auth': {
                'confirmed': check_auth(),
                'http_auth_enabled': config['auth']['http_auth_enabled'],
                'provider': auth_provider
            }
        }

        return resp



"""data-integration-id

Revision ID: d429095b570f
Revises: 1e60096fc817
Create Date: 2022-11-11 14:00:58.386307

"""
import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db as db


# revision identifiers, used by Alembic.
revision = 'd429095b570f'
down_revision = '1e60096fc817'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('data_integration_ref', db.Json(), nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    view_integration = conn.execute('''
        select id from integration where name = 'views'
    ''').fetchone()
    if view_integration is not None:
        views_integration_id = view_integration['id']

        predictors = conn.execute('''
            select id, data_integration_id from predictor
        ''').fetchall()

        for predictor in predictors:
            data_integration_ref = None
            if predictor['data_integration_id'] is not None:
                data_integration_ref = {'type': 'integration', 'id': predictor['data_integration_id']}
            if predictor['data_integration_id'] == views_integration_id:
                data_integration_ref = {'type': 'view'}
            if isinstance(data_integration_ref, dict):
                data_integration_ref = json.dumps(data_integration_ref)
            conn.execute(text('''
                update predictor set data_integration_ref = :data_integration_ref where id = :id
            '''), {
                'data_integration_ref': data_integration_ref,
                'id': predictor['id']
            })

    session.commit()

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('fk_data_integration_id', type_='foreignkey')
        batch_op.drop_column('data_integration_id')

    conn.execute('''
        delete from integration where name = 'views'
    ''')
    session.commit()


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('data_integration_id', sa.INTEGER(), nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    views_integration = db.Integration(
        name='views',
        data={},
        engine='views',
        company_id=None
    )
    session.add(views_integration)
    session.commit()

    predictors = conn.execute('''
        select id, data_integration_ref from predictor
    ''').fetchall()

    for predictor in predictors:
        data_integration_ref = predictor['data_integration_ref']
        if data_integration_ref is None:
            continue
        data_integration_ref = json.loads(data_integration_ref)
        data_integration_id = data_integration_ref.get('id')
        if data_integration_ref['type'] == 'view':
            data_integration_id = views_integration.id

        conn.execute(text('''
            update predictor set data_integration_id = :data_integration_id where id = :id
        '''), {
            'data_integration_id': data_integration_id,
            'id': predictor['id']
        })

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_foreign_key('fk_data_integration_id', 'integration', ['data_integration_id'], ['id'])
        batch_op.drop_column('data_integration_ref')



from collections import OrderedDict
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb.utilities import log
from mindsdb_sql.parser.ast.base import ASTNode
from couchbase.n1ql import N1QLQuery
import pandas as pd

from typing import Optional

from datetime import timedelta

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE



from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


# needed for any cluster connection
from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.exceptions import UnAmbiguousTimeoutException
# needed for options -- cluster, timeout, SQL++ (N1QL) query, etc.
from couchbase.options import (ClusterOptions, ClusterTimeoutOptions,
                               QueryOptions)

from couchbase.exceptions import QueryErrorContext,KeyspaceNotFoundException,CouchbaseException

class CouchbaseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Couchbase statements. 
    """

    name = 'couchbase'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data")
        
        self.scope = self.connection_data.get('scope') or '_default'
        
        self.bucket_name = self.connection_data.get('bucket')
        self.cluster = None
        
        self.is_connected = False

    def connect(self):
        """ Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        """
        if self.is_connected is True:
            return self.cluster
        
        # User Input ends here.
        endpoint = self.connection_data.get('host')
        username = self.connection_data.get('user')
        password = self.connection_data.get('password')

        try:
            timeout = int(self.connection_data.get('timeout'))
        except Exception:
            timeout = 10
        # Connect options - global timeout opts
        timeout_opts = ClusterTimeoutOptions(kv_timeout=timedelta(seconds=timeout))

        auth = PasswordAuthenticator(
            username,
            password,
            # NOTE: If using SSL/TLS, add the certificate path.
            # We strongly reccomend this for production use.
            # cert_path=cert_path
        )

        cluster = Cluster.connect(
            f'couchbase://{endpoint}',
            ClusterOptions(
                authenticator=auth,
                timeout_options=timeout_opts
            )
        )
        
        # # Wait until the cluster is ready for use.
        cluster.wait_until_ready(timedelta(seconds=timeout))
        
        self.is_connected = cluster.connected
        self.cluster = cluster
        return self.cluster

    
    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        self.is_connected = self.cluster.connected
        return
    
    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Couchbase bucket
        :return: success status and error message if error occurs
        """
        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            cluster = self.connect()
            print(self.is_connected)
            result.success = cluster.connected
        except UnAmbiguousTimeoutException as e:
            log.logger.error(f'Error connecting to Couchbase {self.connection_data["bucket"]}, {e}!')
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False
        return result
    
    def native_query(self, query: str) -> Response:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """

        self.connect()
        cluster = self.cluster
        bucket = cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)
        try:
            q = query
            row_iter = cb.query(q)
            data = {}
            # keys = []
            for collection in row_iter:
                # data.append()
                # print(type(collection))
                for collection_name, row in collection.items():
                    if isinstance(row, dict):
                        for k, v in row.items():
                            if data.get(k) is None:
                                data[k] = []
                            data[k].append(v)
                    else:
                        for k, v in collection.items():
                                if data.get(k) is None: 
                                    data[k] = []
                                data[k].append(v)
            if len(data) > 0:
                df = pd.DataFrame(data)
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    df
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
        except CouchbaseException as e:
            print(f'Error: {e.error_context.first_error_message}')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=f'{e.error_context.first_error_message}'
            )
        return response
    
    def query(self, query:ASTNode)-> Response:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        return self.native_query(query.to_string())
    
    def get_tables(self) -> Response:
        """
        Get a list with of collection in database
        """

        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
 
        collections = []

        for _scope in bucket.collections().get_all_scopes():
                for __collections in _scope.collections:
                    collections.append(__collections.name)
        collections_ar = [
            [i] for i in collections
        ]
        
        df = pd.DataFrame(collections_ar, columns=['TABLE_NAME'])
        
        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )
        
        return response

    def get_columns(self, table_name) -> Response:
        """ Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """
        
        response = Response(False)

        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)

        try:
            q = f'SELECT * FROM `{table_name}` limit 1'
            row_iter = cb.query(q)
            # print(row_iter.execute())
            data = []
            for row in row_iter:   
                for k, v in row[table_name].items():
                    data.append([k, type(v).__name__])
            df = pd.DataFrame(data, columns=['Field', 'Type'])
            response = Response(
                RESPONSE_TYPE.TABLE,
                df
            )
        except KeyspaceNotFoundException as e:
            print(f'Error: {e.error_context.first_error_message}')
            response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=f'Error: {e.error_context.first_error_message}'
                )
                
        return response



connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Couchbase server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Couchbase server.'
    },
    bucket={
        'type': ARG_TYPE.STR,
        'description': 'The database/bucket name to use when connecting with the Couchbase server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': '--your-instance--.dp.cloud.couchbase.com or IP address of the Couchbase server.'
    },
    scope={
        'type': ARG_TYPE.STR,
        'description': 'The scope use in the query context in Couchbase server. If blank, scope will be "_default".'
    }
)
connection_args_example = OrderedDict(
    host='127.0.0.1',
    user='root',
    password='password',
    bucket='bucket'
)



import os
# from .executor import Executor
from .executor_commands import ExecuteCommands
from .data_types import *
# from .executor_client import ExecutorClient
from .executor_client_factory import ExecutorClient


# In case of Executor Service need to use ExecutorClient instead of Executor
# so here we hide the difference between two objects
# if os.environ.get("MINDSDB_EXECUTOR_URL"):
Executor = ExecutorClient



import shopify
import pandas as pd
from typing import Text, List, Dict, Any

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.handlers.utilities.query_utilities.insert_query_utilities import INSERTQueryParser

from mindsdb.utilities.log import get_log

logger = get_log("integrations.shopify_handler")


class ProductsTable(APITable):
    """The Shopify Products Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /products" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Products matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'products',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            products_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        products_df = select_statement_executor.execute_query()

        return products_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_products(limit=1)).columns.tolist()

    def get_products(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        products = shopify.Product.find(**kwargs)
        return [product.to_dict() for product in products]


class CustomersTable(APITable):
    """The Shopify Customers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /customers" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Customers matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'customers',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            customers_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        customers_df = select_statement_executor.execute_query()

        return customers_df

    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Shopify "POST /customers" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['first_name', 'last_name', 'email', 'phone', 'tags', 'currency'],
            mandatory_columns=['first_name', 'last_name', 'email', 'phone'],
            all_mandatory=False
        )
        customer_data = insert_statement_parser.parse_query()
        self.create_customers(customer_data)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_customers(limit=1)).columns.tolist()

    def get_customers(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        customers = shopify.Customer.find(**kwargs)
        return [customer.to_dict() for customer in customers]

    def create_customers(self, customer_data: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for customer in customer_data:
            created_customer = shopify.Customer.create(customer)
            if 'id' not in created_customer.to_dict():
                raise Exception('Customer creation failed')
            else:
                logger.info(f'Customer {created_customer.to_dict()["id"]} created')


class OrdersTable(APITable):
    """The Shopify Orders Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /orders" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Orders matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'orders',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        orders_df = pd.json_normalize(self.get_orders(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            orders_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        orders_df = select_statement_executor.execute_query()

        return orders_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_orders(limit=1)).columns.tolist()

    def get_orders(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        orders = shopify.Order.find(**kwargs)
        return [order.to_dict() for order in orders]


# https://stackoverflow.com/questions/29298346/xmpp-sasl-scram-sha1-authentication
# https://tools.ietf.org/html/rfc5802
# https://tools.ietf.org/html/rfc7677

import base64
import hashlib
import hmac
import os

from pymongo.auth import _password_digest, _xor, saslprep


class Scram():
    ''' implementation of server-side SCRAM-SHA-1 and SCRAM-SHA-256 auth for mongodb
    '''

    def __init__(self, method='sha1', get_salted_password=None):
        self.get_salted_password = get_salted_password
        self.snonce = base64.b64encode(os.urandom(24)).decode()
        self.iterations = 4096
        self.messages = []

        if method == 'sha1':
            self.method_str = 'sha1'
            self.method_func = hashlib.sha1
            self.salt = base64.b64encode(os.urandom(16))
        elif method == 'sha256':
            self.method_str = 'sha256'
            self.method_func = hashlib.sha256
            self.salt = base64.b64encode(os.urandom(28))

    def process_client_first_message(self, payload):
        payload = payload[3:]
        payload_parts = self._split_payload(payload)
        self.client_user = payload_parts['n']

        if self.get_salted_password is not None:
            salt_bytes, self.salted_password = self.get_salted_password(self.client_user, self.method_str)
            self.salt = base64.b64encode(salt_bytes)
        else:
            self.salted_password = self.salt_password()

        self.unonce = payload_parts['r']
        self.messages.append(payload)

        responce_msg = f"r={self.unonce}{self.snonce},s={self.salt.decode()},i={self.iterations}"
        self.messages.append(responce_msg)
        return responce_msg

    def process_client_second_message(self, payload):
        self.messages.append(payload[:payload.rfind(',p=')])    # without 'p' part

        messages = ','.join(self.messages)
        server_key = self._hmac(self.salted_password, b'Server Key')
        server_signature = self._hmac(server_key, messages.encode('utf-8'))

        client_key = self._hmac(self.salted_password, b'Client Key')
        stored_key = self.method_func(client_key).digest()
        client_signature = self._hmac(stored_key, messages.encode('utf-8'))
        expected_client_proof = base64.b64encode(_xor(client_key, client_signature)).decode()

        income_client_proof = payload[payload.rfind(',p=') + 3:]

        if expected_client_proof != income_client_proof:
            raise Exception('wrong password')

        return f'v={base64.b64encode(server_signature).decode()}'

    def _hmac(self, key, msg):
        return hmac.new(key, msg, digestmod=self.method_func).digest()

    def salt_password(self, user, password):
        if self.method_str == 'sha1':
            password = _password_digest(user, password).encode("utf-8")
        elif self.method_str == 'sha256':
            password = saslprep(password).encode("utf-8")

        return hashlib.pbkdf2_hmac(
            self.method_str, password, base64.b64decode(self.salt), self.iterations
        )

    def _split_payload(self, payload):
        parts = {}
        for part in [x for x in payload.split(',') if len(x) > 0]:
            parts[part[0]] = part[2:]
        return parts



from http import HTTPStatus

from flask import request
from flask_restx import Resource
from sqlalchemy.exc import NoResultFound

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.api.http.utils import http_error
from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor


@ns_conf.route('/<project_name>/chatbots')
class ChatBotsResource(Resource):
    @ns_conf.doc('list_chatbots')
    def get(self, project_name):
        ''' List all chatbots '''
        chatbot_controller = ChatBotController()
        try:
            all_bots = chatbot_controller.get_chatbots(project_name)
        except NoResultFound:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist')
        return [b.as_dict() for b in all_bots]

    @ns_conf.doc('create_chatbot')
    def post(self, project_name):
        '''Create a chatbot'''

        # Check for required parameters.
        if 'chatbot' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "chatbot" parameter in POST body'
            )

        chatbot = request.json['chatbot']
        if 'name' not in chatbot:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing field',
                'Missing "name" field for chatbot'
            )
        if 'model_name' not in chatbot:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing field',
                'Missing "model_name" field for chatbot'
            )
        if 'database_id' not in chatbot and 'chat_engine' not in chatbot:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing field',
                'Must provide either "database_id" or "chat_engine" when creating a chatbot'
            )

        name = chatbot['name']
        model_name = chatbot['model_name']
        database_id = chatbot.get('database_id', None)
        chat_engine = chatbot.get('chat_engine', None)
        is_running = chatbot.get('is_running', False)
        params = chatbot.get('params', {})

        chatbot_controller = ChatBotController()
        session_controller = SessionController()

        # Chatbot can't already exist.
        try:
            existing_chatbot = chatbot_controller.get_chatbot(name, project_name=project_name)
            if existing_chatbot is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    'Chatbot already exists',
                    f'Chatbot with name {name} already exists. Please use a different name'
                )
        except NoResultFound:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        # Model needs to exist.
        model_name_no_version, version = Predictor.get_name_and_version(model_name)
        try:
            session_controller.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {model_name} not found')

        created_chatbot = chatbot_controller.add_chatbot(
            name,
            project_name,
            model_name,
            database_id=database_id,
            chat_engine=chat_engine,
            is_running=is_running,
            params=params
        )
        return created_chatbot.as_dict(), HTTPStatus.CREATED


@ns_conf.route('/<project_name>/chatbots/<chatbot_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('chatbot_name', 'Name of the chatbot')
class ChatBotResource(Resource):
    @ns_conf.doc('get_chatbot')
    def get(self, project_name, chatbot_name):
        '''Gets a chatbot by name'''
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Chatbot not found',
                    f'Chatbot with name {chatbot_name} does not exist'
                )
            return existing_chatbot.as_dict()
        except NoResultFound:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

    @ns_conf.doc('update_chatbot')
    def put(self, project_name, chatbot_name):
        '''Updates a chatbot by name, creating one if it doesn't exist'''

        # Check for required parameters.
        if 'chatbot' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "chatbot" parameter in POST body'
            )
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
        except NoResultFound:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        chatbot = request.json['chatbot']
        name = chatbot.get('name', None)
        model_name = chatbot.get('model_name', None)
        database_id = chatbot.get('database_id', None)
        chat_engine = chatbot.get('chat_engine', None)
        is_running = chatbot.get('is_running', None)
        params = chatbot.get('params', None)

        # Model needs to exist.
        if model_name is not None:
            session_controller = SessionController()
            model_name_no_version, version = Predictor.get_name_and_version(model_name)
            try:
                session_controller.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
            except PredictorRecordNotFound:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Model not found',
                    f'Model with name {model_name} not found')

        # Chatbot must not exist with new name.
        if name is not None and name != chatbot_name:
            chatbot_with_new_name = chatbot_controller.get_chatbot(name, project_name=project_name)
            if chatbot_with_new_name is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    'Chatbot already exists',
                    f'Chatbot with name {name} already exists. Please choose a different one.'
                )

        if existing_chatbot is None:
            # Create
            if name is None:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing field',
                    'Missing "name" field for chatbot'
                )
            if model_name is None:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing field',
                    'Missing "model_name" field for chatbot'
                )
            if database_id is None and chat_engine is None:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing field',
                    'Must provide either "database_id" or "chat_engine" when creating a chatbot'
                )

            created_chatbot = chatbot_controller.add_chatbot(
                name,
                project_name,
                model_name,
                database_id=database_id,
                chat_engine=chat_engine,
                is_running=is_running,
                params=params if params else {}
            )
            return created_chatbot.as_dict(), HTTPStatus.CREATED

        # Update
        updated_chatbot = chatbot_controller.update_chatbot(
            chatbot_name,
            project_name=project_name,
            name=name,
            model_name=model_name,
            database_id=database_id,
            chat_engine=chat_engine,
            is_running=is_running,
            params=params
        )
        return updated_chatbot.as_dict()

    @ns_conf.doc('delete_chatbot')
    def delete(self, project_name, chatbot_name):
        '''Deletes a chatbot by name'''
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Chatbot not found',
                    f'Chatbot with name {chatbot_name} does not exist'
                )
        except NoResultFound:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        chatbot_controller.delete_chatbot(chatbot_name, project_name=project_name)
        return '', HTTPStatus.NO_CONTENT



import datetime as dt
import pytz
import re


def parse_local_date(date_str: str) -> dt.datetime:
    """Parses common date string formats to local datetime objects."""
    if isinstance(date_str, dt.datetime):
        return date_str
    date_formats = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d']
    date = None
    for date_format in date_formats:
        try:
            date = dt.datetime.strptime(date_str, date_format)
        except ValueError:
            pass
    if date is None:
        raise ValueError(f"Can't parse date: {date_str}")
    return date


def parse_utc_date(date_str: str) -> dt.datetime:
    """Parses common date string formats to UTC datetime objects."""
    date = parse_local_date(date_str)
    return date.astimezone(pytz.utc)


def utc_date_str_to_timestamp_ms(date_str: str) -> int:
    """Parses common date string formats into ms since the Unix epoch in UTC."""
    date = parse_local_date(date_str)
    # `timestamp` method doesn't work as expected unless we replace the timezone info this way.
    date = date.replace(tzinfo=pytz.UTC)
    return int(date.timestamp() * 1000)


def interval_str_to_duration_ms(interval_str: str) -> int:
    """Parses interval strings into how long they represent in ms.
    Supported intervals:
      - seconds (e.g. 1s)
      - minutes (e.g. 5m)
      - hours (e.g. 1h)
      - days (e.g. 5d)
      - weeks (e.g. 1w)
    """
    duration_match = re.search(r'^\d+', interval_str)
    time_unit_match = re.search('[smhdw]', interval_str)
    if not duration_match or not time_unit_match:
        raise ValueError('Invalid interval {}'.format(interval_str))
    duration = int(duration_match.group())
    time_unit = time_unit_match.group()
    if time_unit == 's':
        return duration * 1000
    if time_unit == 'm':
        return duration * 1000 * 60
    if time_unit == 'h':
        return duration * 1000 * 60 * 60
    if time_unit == 'd':
        return duration * 1000 * 60 * 60 * 24
    if time_unit == 'w':
        return duration * 1000 * 60 * 60 * 24 * 7



import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from typing import List, Tuple
import requests


class StoriesTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the stories table and return it as a pandas DataFrame.
        Args:
            query (ast.Select): The SQL query to be executed.
        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """
        hn_handler = self.handler

        # Extract the limit value from the SQL query, if it exists
        limit = None
        if query.limit is not None:
            limit = query.limit.value

        # Call the Hacker News API to get the top stories
        url = f'{hn_handler.base_url}/topstories.json'
        response = requests.get(url)
        data = response.json()

        # Fetch the details of the top stories, up to the specified limit
        stories_data = []
        for story_id in data[:limit]:
            url = f'{hn_handler.base_url}/item/{story_id}.json'
            response = requests.get(url)
            story_data = response.json()
            stories_data.append(story_data)

        # Create a DataFrame from the fetched data
        df = pd.DataFrame(
            stories_data, 
            columns=['id', 'time', 'title', 'url', 'score', 'descendants']
            )

        # Apply any WHERE clauses in the SQL query to the DataFrame
        conditions = extract_comparison_conditions(query.where)
        for condition in conditions:
            if condition[0] == '=' and condition[1] == 'id':
                df = df[df['id'] == int(condition[2])]
            elif condition[0] == '>' and condition[1] == 'time':
                timestamp = int(condition[2])
                df = df[df['time'] > timestamp]

        # Filter the columns in the DataFrame according to the SQL query
        self.filter_columns(df, query)

        return df


    def get_columns(self):
        """Get the list of column names for the stories table.
        Returns:
            list: A list of column names for the stories table.
        """
        return ['id', 'time', 'title', 'url', 'score', 'descendants']

    def filter_columns(self, df, query):
        """Filter the columns in the DataFrame according to the SQL query.
        Args:
            df (pandas.DataFrame): The DataFrame to filter.
            query (ast.Select): The SQL query to apply to the DataFrame.
        """
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.value)
        df = df[columns]
        return df


class CommentsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the comments table and return it as a pandas DataFrame.
        Args:
            query (ast.Select): The SQL query to be executed.
        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """
        hn_handler = self.handler

        # Get the limit value from the SQL query, if it exists
        limit = None
        if query.limit is not None:
            limit = query.limit.value

        # Get the item ID from the SQL query
        item_id = None
        conditions = extract_comparison_conditions(query.where)
        for condition in conditions:
            if condition[0] == '=' and condition[1] == 'item_id':
                item_id = condition[2]

        if item_id is None:
            raise ValueError('Item ID is missing in the SQL query')

        # Call the Hacker News API to get the comments for the specified item
        comments_df = hn_handler.call_hackernews_api('get_comments', params={'item_id': item_id})

        # Fill NaN values with 'deleted'
        comments_df = comments_df.fillna('deleted')
        # Filter the columns to those specified in the SQL query
        self.filter_columns(comments_df, query)

        # Limit the number of results if necessary
        if limit is not None:
            comments_df = comments_df.head(limit)

        return comments_df
    
    
    def get_columns(self) -> List[str]:
        """Get the list of column names for the comments table.
        Returns:
            list: A list of column names for the comments table.
        """
        return [
            'id',
            'by',
            'parent',
            'text',
            'time',
            'type',
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None) -> None:
        """Filter the columns of a DataFrame to those specified in an SQL query.
        Args:
            result (pandas.DataFrame): The DataFrame to filter.
            query (ast.Select): The SQL query containing the column names to filter on.
        """
        if query is None:
            return

        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                return
            elif isinstance(target, ast.Identifier):
                columns.append(target.value)

        if len(columns) > 0:
            result = result[columns]



import os
import json
import datetime
from typing import Dict

import numpy as np
from sqlalchemy import create_engine, types, UniqueConstraint
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Boolean, Index, text
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy import JSON
from sqlalchemy.exc import OperationalError

Base = declarative_base()
session, engine = None, None


def init(connection_str: str = None):
    global Base, session, engine
    if connection_str is None:
        connection_str = os.environ['MINDSDB_DB_CON']
    if connection_str.startswith('sqlite:'):
        engine = create_engine(connection_str, echo=False)
    else:
        engine = create_engine(connection_str, convert_unicode=True, pool_size=30, max_overflow=200, echo=False)
    session = scoped_session(sessionmaker(bind=engine, autoflush=True))
    Base.query = session.query_property()


def serializable_insert(record: Base, try_count: int = 100):
    """ Do serializeble insert. If fail - repeat it {try_count} times.

        Args:
            record (Base): sqlalchey record to insert
            try_count (int): count of tryes to insert record
    """
    commited = False
    while not commited:
        session.connection(
            execution_options={'isolation_level': 'SERIALIZABLE'}
        )
        if engine.name == 'postgresql':
            session.execute(text('LOCK TABLE PREDICTOR IN EXCLUSIVE MODE'))
        session.add(record)
        try:
            session.commit()
        except OperationalError:
            # catch 'SerializationFailure' (it should be in str(e), but it may depend on engine)
            session.rollback()
            try_count += -1
            if try_count == 0:
                raise
        else:
            commited = True


# Source: https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Array(types.TypeDecorator):
    ''' Float Type that replaces commas with  dots on input '''
    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        if isinstance(value, str):
            return value
        elif value is None:
            return value
        else:
            return ',|,|,'.join(value)

    def process_result_value(self, value, dialect):  # select
        return value.split(',|,|,') if value is not None else None


class Json(types.TypeDecorator):
    ''' Float Type that replaces commas with  dots on input '''
    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        return json.dumps(value, cls=NumpyEncoder) if value is not None else None

    def process_result_value(self, value, dialect):  # select
        if isinstance(value, dict):
            return value
        return json.loads(value) if value is not None else None


class Semaphor(Base):
    __tablename__ = 'semaphor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    entity_type = Column('entity_type', String)
    entity_id = Column('entity_id', Integer)
    action = Column(String)
    company_id = Column(Integer)
    __table_args__ = (
        UniqueConstraint('entity_type', 'entity_id', name='uniq_const'),
    )


class PREDICTOR_STATUS:
    __slots__ = ()
    COMPLETE = 'complete'
    TRAINING = 'training'
    FINETUNING = 'finetuning'
    GENERATING = 'generating'
    ERROR = 'error'
    VALIDATION = 'validation'
    DELETED = 'deleted'  # TODO remove it?


PREDICTOR_STATUS = PREDICTOR_STATUS()


class Predictor(Base):
    __tablename__ = 'predictor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    deleted_at = Column(DateTime)
    name = Column(String)
    data = Column(Json)  # A JSON -- should be everything returned by `get_model_data`, I think
    to_predict = Column(Array)
    company_id = Column(Integer)
    mindsdb_version = Column(String)
    native_version = Column(String)
    integration_id = Column(ForeignKey('integration.id', name='fk_integration_id'))
    data_integration_ref = Column(Json)
    fetch_data_query = Column(String)
    is_custom = Column(Boolean)
    learn_args = Column(Json)
    update_status = Column(String, default='up_to_date')
    status = Column(String)
    active = Column(Boolean, default=True)
    training_data_columns_count = Column(Integer)
    training_data_rows_count = Column(Integer)
    training_start_at = Column(DateTime)
    training_stop_at = Column(DateTime)
    label = Column(String, nullable=True)
    version = Column(Integer, default=1)

    code = Column(String, nullable=True)
    lightwood_version = Column(String, nullable=True)
    dtype_dict = Column(Json, nullable=True)
    project_id = Column(Integer, ForeignKey('project.id', name='fk_project_id'), nullable=False)
    training_phase_current = Column(Integer)
    training_phase_total = Column(Integer)
    training_phase_name = Column(String)

    @staticmethod
    def get_name_and_version(full_name):
        name_no_version = full_name
        version = None
        parts = full_name.split('.')
        if len(parts) > 1 and parts[-1].isdigit():
            version = int(parts[-1])
            name_no_version = '.'.join(parts[:-1])
        return name_no_version, version


class Project(Base):
    __tablename__ = 'project'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    deleted_at = Column(DateTime)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    __table_args__ = (
        UniqueConstraint('name', 'company_id', name='unique_project_name_company_id'),
    )


class Log(Base):
    __tablename__ = 'log'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    log_type = Column(String)  # log, info, warning, traceback etc
    source = Column(String)  # file + line
    company_id = Column(Integer)
    payload = Column(String)
    created_at_index = Index("some_index", "created_at_index")


class Integration(Base):
    __tablename__ = 'integration'
    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String, nullable=False)
    engine = Column(String, nullable=False)
    data = Column(Json)
    company_id = Column(Integer)
    __table_args__ = (
        UniqueConstraint('name', 'company_id', name='unique_integration_name_company_id'),
    )


class File(Base):
    __tablename__ = 'file'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    source_file_path = Column(String, nullable=False)
    file_path = Column(String, nullable=False)
    row_count = Column(Integer, nullable=False)
    columns = Column(Json, nullable=False)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    __table_args__ = (
        UniqueConstraint('name', 'company_id', name='unique_file_name_company_id'),
    )


class View(Base):
    __tablename__ = 'view'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    query = Column(String, nullable=False)
    project_id = Column(Integer, ForeignKey('project.id', name='fk_project_id'), nullable=False)
    __table_args__ = (
        UniqueConstraint('name', 'company_id', name='unique_view_name_company_id'),
    )


class JsonStorage(Base):
    __tablename__ = 'json_storage'
    id = Column(Integer, primary_key=True)
    resource_group = Column(String)
    resource_id = Column(Integer)
    name = Column(String)
    content = Column(JSON)
    company_id = Column(Integer)


class Jobs(Base):
    __tablename__ = 'jobs'
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)
    user_class = Column(Integer, nullable=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)
    query_str = Column(String, nullable=False)
    start_at = Column(DateTime, default=datetime.datetime.now)
    end_at = Column(DateTime)
    next_run_at = Column(DateTime)
    schedule_str = Column(String)

    deleted_at = Column(DateTime)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)


class JobsHistory(Base):
    __tablename__ = 'jobs_history'
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)

    job_id = Column(Integer)

    query_str = Column(String)
    start_at = Column(DateTime)
    end_at = Column(DateTime)

    error = Column(String)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(DateTime, default=datetime.datetime.now)

    __table_args__ = (
        UniqueConstraint('job_id', 'start_at', name='uniq_job_history_job_id_start'),
    )


class ChatBots(Base):
    __tablename__ = 'chat_bots'
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)
    user_class = Column(Integer, nullable=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)

    model_name = Column(String, nullable=False)
    # If database_id is set we use an API Handler to poll chat messages.
    database_id = Column(Integer)
    # If chat_engine is set we use a RealtimeChatHandler to subscribe to chat messages.
    # TODO(tmichaeldb): Consolidate existing polling logic and realtime chat logic together.
    chat_engine = Column(String)
    params = Column(JSON)

    is_running = Column(Boolean, default=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)

    def as_dict(self) -> Dict:
        return {
            'id': self.id,
            'name': self.name,
            'project_id': self.project_id,
            'model_name': self.model_name,
            'chat_engine': self.chat_engine,
            'params': self.params,
            'is_running': self.is_running,
            'created_at': self.created_at
        }


class ChatBotsHistory(Base):
    __tablename__ = 'chat_bots_history'
    id = Column(Integer, primary_key=True)
    chat_bot_id = Column(Integer)
    type = Column(String)
    text = Column(String)
    user = Column(String)
    destination = Column(String)
    sent_at = Column(DateTime, default=datetime.datetime.now)
    error = Column(String)



import os
from typing import Optional, Dict

import openai
import pandas as pd
from langchain.llms import OpenAI
import llama_index
from llama_index.readers.schema.base import Document
from llama_index import SimpleWebPageReader, QuestionAnswerPrompt
from llama_index import ServiceContext, StorageContext, load_index_from_storage
from llama_index import LLMPredictor, OpenAIEmbedding
from llama_index.indices.vector_store.base import VectorStore

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities.config import Config


class LlamaIndexHandler(BaseMLEngine):
    """ Integration with the LlamaIndex data framework for LLM applications. """
    name = 'llama_index'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.default_index_class = 'GPTVectorStoreIndex'
        self.supported_index_class = ['GPTVectorStoreIndex']
        self.default_reader = 'DFReader'
        self.supported_reader = ['DFReader', 'SimpleWebPageReader']

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("LlamaIndex engine requires a USING clause! Refer to its documentation for more details.")

        if 'prompt_template' in args['using']:
            self._validate_prompt_template(args['using']['prompt_template'])

        if 'index_class' not in args['using']:
            args['using']['index_class'] = self.default_index_class
        elif args['using']['index_class'] not in self.supported_index_class:
            raise Exception(f"Invalid index class argument. Please use one of {self.supported_index_class}")

        if 'reader' not in args['using']:
            args['using']['reader'] = self.default_reader
        elif args['using']['reader'] not in self.supported_reader:
            raise Exception(f"Invalid operation mode. Please use one of {self.supported_reader}")

        if args['using']['reader'] == 'DFReader':
            dstrs = df.apply(lambda x: ', '.join([f'{col}: {str(entry)}' for col, entry in zip(df.columns, x)]), axis=1)
            reader = list(map(lambda x: Document(x), dstrs.tolist()))

        elif args['using']['reader'] == 'SimpleWebPageReader':
            if 'source_url_link' not in args['using']:
                raise Exception("SimpleWebPageReader requires a `source_url_link` parameter. Refer to LlamaIndex documentation for more details.")  # noqa

            reader = SimpleWebPageReader(html_to_text=True).load_data([args['using']['source_url_link']])

        else:
            raise Exception(f"Invalid operation mode. Please use one of {self.supported_reader}.")

        self.model_storage.json_set('args', args)
        index = self._setup_index(reader)
        path = self.model_storage.fileStorage.get_path('./')
        index.storage_context.persist(persist_dir=path)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        input_column = args['using'].get('input_column', None)
        engine_kwargs = {}

        prompt_template = args['using'].get('prompt_template', args.get('prompt_template', None))
        if prompt_template is not None:
            self._validate_prompt_template(prompt_template)
            engine_kwargs['text_qa_template'] = QuestionAnswerPrompt(prompt_template)

        if input_column is None:
            raise Exception(f'`input_column` must be provided at model creation time or through USING clause when predicting. Please try again.')  # noqa

        if input_column not in df.columns:
            raise Exception(f'Column "{input_column}" not found in input data! Please try again.')

        index_path = self.model_storage.fileStorage.get_path('./')
        storage_context = StorageContext.from_defaults(persist_dir=index_path)
        service_context = self._get_service_context()
        index = load_index_from_storage(storage_context, service_context=service_context)
        query_engine = index.as_query_engine(**engine_kwargs)

        questions = df[input_column]
        results = []

        for question in questions:
            query_results = query_engine.query(question)  # TODO: provide extra_info in explain_target col
            results.append(query_results.response)

        result_df = pd.DataFrame({'question': questions, args['target']: results})  # result_df['answer'].tolist()
        return result_df

    def _get_service_context(self):
        args = self.model_storage.json_get('args')
        openai_api_key = self._get_llama_index_api_key(args['using'])
        openai.api_key = openai_api_key  # TODO: shouldn't have to do this! bug?
        llm = OpenAI(openai_api_key=openai_api_key)  # TODO: all usual params should go here
        embed_model = OpenAIEmbedding(openai_api_key=openai_api_key)
        service_context = ServiceContext.from_defaults(
            llm_predictor=LLMPredictor(llm=llm),
            embed_model=embed_model
        )
        return service_context
    
    def _setup_index(self, documents):
        args = self.model_storage.json_get('args')
        indexer: VectorStore = getattr(llama_index, args['using']['index_class'])
        index = indexer.from_documents(documents, service_context=self._get_service_context())

        return index

    def _validate_prompt_template(self, prompt_template: str):
        if '{context_str}' not in prompt_template or '{query_str}' not in prompt_template:
            raise Exception("Provided prompt template is invalid, missing one of `{context_str}` or `{query_str}`. Please ensure both placeholders are present and try again.")  # noqa

    def _get_llama_index_api_key(self, args, strict=True):
        """
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. OPENAI_API_KEY env variable
            4. llama_index.OPENAI_API_KEY setting in config.json

        Note: method is not case sensitive.
        """
        key = 'OPENAI_API_KEY'
        for k in key, key.lower():
            # 1
            if args.get(k):
                return args[k]
            # 2
            connection_args = self.engine_storage.get_connection_args()
            if k in connection_args:
                return connection_args[k]
            # 3
            api_key = os.getenv(k)
            if api_key is not None:
                return api_key
            # 4
            config = Config()
            openai_cfg = config.get('llama_index', {})
            if k in openai_cfg:
                return openai_cfg[k]

        if strict:
            raise Exception(f'Missing API key "{k}". Either re-create this ML_ENGINE specifying the `{k}` parameter, or re-create this model and pass the API key with `USING` syntax.')  # noqa



import shopify

from mindsdb.integrations.handlers.shopify_handler.shopify_tables import ProductsTable, CustomersTable, OrdersTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql import parse_sql


class ShopifyHandler(APIHandler):
    """
    The Shopify handler implementation.
    """

    name = 'shopify'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        products_data = ProductsTable(self)
        self._register_table("products", products_data)

        customers_data = CustomersTable(self)
        self._register_table("customers", customers_data)

        orders_data = OrdersTable(self)
        self._register_table("orders", orders_data)

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        api_session = shopify.Session(self.connection_data['shop_url'], '2021-10', self.connection_data['access_token'])

        self.connection = api_session

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            api_session = self.connect()
            shopify.ShopifyResource.activate_session(api_session)
            shopify.Shop.current()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Shopify!')
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


import os
from datetime import datetime as datetime, timezone
import ast
from typing import List
import pandas as pd
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.web.slack_response import SlackResponse

from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb_sql.parser import ast
from mindsdb_sql.parser.ast import ASTNode, Update, Delete
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.integrations.libs.api_handler import APIHandler, APITable, FuncParser
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

class SlackChannelsTable(APITable):
    def __init__(self, handler):
        """
        Checks the connection is active
        """
        super().__init__(handler)
        self.client = WebClient(token=self.handler.connection_args['token'])

    def select(self, query: ast.Select) -> Response:
        """
        Retrieves the data from the channel using SlackAPI

        Args:
            channel_name

        Returns:
            conversation_history
        """
        # override the default function
        def parse_utc_date(date_str):
            date_obj = datetime.fromisoformat(date_str).replace(tzinfo=timezone.utc)
            return date_obj.strftime('%Y-%m-%d %H:%M:%S')

        # Get the channels list and ids
        channels = self.client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}
        
        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)
        channel_name = conditions[0][2];
        filters = []
        params = {}
        order_by_conditions = {}
        
        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if arg1 == 'channel':
                if arg2 in channel_ids:
                    params['channel'] = channel_ids[arg2]
                else:
                    raise ValueError(f"Channel '{arg2}' not found")

            elif arg1 == 'limit':
                if op == '=': 
                    params['limit'] = int(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
                
            elif arg1 == 'message_created_at':
                date = parse_utc_date(arg2)
                if op == '>':
                    params['start_time'] = date
                elif op == '<':
                    params['end_time'] = date
                else:
                    raise NotImplementedError

            else:
                filters.append([op, arg1, arg2])

        if query.limit:
            params['limit'] = int(query.limit.value)

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[1] == "messages":
                    order_by_conditions["columns"].append("messages")

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        # Retrieve the conversation history
        try:
            result = self.client.conversations_history(channel=params['channel'])
            conversation_history = result["messages"]
        except SlackApiError as e:
            log.logger.error("Error creating conversation: {}".format(e))

        # Get columns for the query and convert SlackResponse object to pandas DataFrame
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        # convert SlackResponse object to pandas DataFrame
        result = pd.DataFrame(result['messages'], columns=columns)
            
        # add absent columns
        for col in set(columns) & set(result.columns) ^ set(columns):
            result[col] = None

        # filter by columns
        columns = [target.parts[-1].lower() for target in query.targets if isinstance(target, ast.Identifier)]
        result = result[columns]

        # Append the history, timestamp, datetime to the response
        response_history = []
        response_ts = []
        response_datetime = []
        response_datetime_timestamp = []
        
        # if we have 'message_created_at' attribute, then execute this
        if 'start_time' in params or 'end_time' in params:
            # for every message convert ts to str and ts to DateTime object and message to the response object
            for message in conversation_history:
                timestamp = message['ts']
                float_ts = float(message['ts'])
                datetime_ts = datetime.fromtimestamp(float_ts, tz=timezone.utc)
                datetime_ts = datetime_ts.strftime('%Y-%m-%d %H:%M:%S')
                response_datetime.append(datetime_ts)
                response_ts.append(timestamp)
                
                # if '>' operator store message in response_datetime_timestamp
                if 'start_time' in params and datetime_ts > params['start_time']:
                    response_datetime_timestamp.append(message['text'])
                # if '<' operator store message in response_datetime_timestamp
                elif 'end_time' in params and datetime_ts < params['end_time']:
                    response_datetime_timestamp.append(message['text'])
                # else store it in response_history
                else:
                    response_history.append(message['text'])
                    
        # else parse message and return as a response object
        else:
            # for every message convert ts to str and ts to DateTime object and message to the response object
            for message in conversation_history:
                timestamp = message['ts']
                float_ts = float(message['ts'])
                datetime_ts = datetime.fromtimestamp(float_ts, tz=timezone.utc)
                datetime_ts = datetime_ts.strftime('%Y-%m-%d %H:%M:%S')
                response_datetime.append(datetime_ts)
                response_ts.append(timestamp)
                response_history.append(message['text'])
                           
        result['channel'] = channel_name
        # convert response_datetime_timestamp to Dataframe object in order to return
        if 'start_time' in params or 'end_time' in params:
            filtered_df = pd.DataFrame(response_datetime_timestamp, columns=result.columns)
            result['messages'] = filtered_df
        else:
            result['messages'] = response_history
        
        result['message_created_at'] = response_datetime
        result['ts'] = response_ts
        
        # Sort the data based on order_by_conditions
        if len(order_by_conditions.get("columns", [])) > 0:
            result = result.sort_values(
                by=order_by_conditions["columns"],
                ascending=order_by_conditions["ascending"],
            )

        # Limit the result based on the query limit
        if query.limit:
            result = result.head(query.limit.value)
            
        # Alias the target column based on the query
        for target in query.targets:
            if target.alias:
                result.rename(columns={target.parts[-1]: str(target.alias)}, inplace=True)
        
        # Remove null rows from the result
        result = result[result['messages'].notnull()]
        return result
        
    def get_columns(self):
        """
        Returns columns from the SlackAPI
        """

        return [
            'ts',
            'text',
            'message_created_at',
            'user',
            'channel',
            'reactions',
            'attachments',
            'thread_ts',
            'reply_count',
            'reply_users_count',
            'latest_reply',
            'subtype',
            'hidden',
        ]

    def insert(self, query):
        """
        Inserts the message in the Slack Channel

        Args:
            channel_name
            message
        """
    
        # get column names and values from the query
        columns = [col.name for col in query.columns]
        for row in query.values:
            params = dict(zip(columns, row))

            # check if required parameters are provided
            if 'channel' not in params or 'message' not in params:
                raise Exception("To insert data into Slack, you need to provide the 'channel' and 'message' parameters.")

            # post message to Slack channel
            try:
                response = self.client.chat_postMessage(
                    channel=params['channel'],
                    text=params['message']
                )
            except SlackApiError as e:
                raise Exception(f"Error posting message to Slack channel '{params['channel']}': {e.response['error']}")
            
            inserted_id = response['ts']
            params['ts'] = inserted_id

    def update(self, query: ast.Update):
        """
        Updates the message in the Slack Channel

        Args:
            updated message
            channel_name
            ts  [TimeStamp -> Can be found by running select command, the entire result will be printed in the terminal]
        """
        
        # Get the channels list and ids
        channels = self.client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}
        
        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)
        
        filters = []
        params = {}
        
        keys = list(query.update_columns.keys())
        
        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if arg1 == 'channel':
                if arg2 in channel_ids:
                    params['channel'] = channel_ids[arg2]
                else:
                    raise ValueError(f"Channel '{arg2}' not found")
           
            if keys[0] == 'message':
                params['message'] = str(query.update_columns['message'])
            else:
                raise ValueError(f"Message '{arg2}' not found")

            if arg1 == 'ts':
                if op == '=': 
                    params['ts'] = float(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
            else:
                filters.append([op, arg1, arg2])

        # check if required parameters are provided
        if 'channel' not in params or 'ts' not in params or 'message' not in params:
            raise Exception("To update a message in Slack, you need to provide the 'channel', 'ts', and 'message' parameters.")

        # update message in Slack channel
        try:
            response = self.client.chat_update(
                channel=params['channel'],
                ts=params['ts'],
                text=params['message']
            )
        except SlackApiError as e:
            raise Exception(f"Error updating message in Slack channel '{params['channel']}' with timestamp '{params['ts']}' and message '{params['message']}': {e.response['error']}")
        
        return response
    
    def delete(self, query: ast.Delete):
        """
        Deletes the message in the Slack Channel

        Args:
            channel_name
            ts  [TimeStamp -> Can be found by running select command, the entire result will be printed in the terminal]
        """
        # Get the channels list and ids
        channels = self.client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}
        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)
        
        filters = []
        params = {}
        
        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if arg1 == 'channel':
                if arg2 in channel_ids:
                    params['channel'] = channel_ids[arg2]
                else:
                    raise ValueError(f"Channel '{arg2}' not found")

            if arg1 == 'ts':
                if op == '=': 
                    params['ts'] = float(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')

            else:
                filters.append([op, arg1, arg2])

        # check if required parameters are provided
        if 'channel' not in params or 'ts' not in params:
            raise Exception("To delete a message from Slack, you need to provide the 'channel' and 'ts' parameters.")

        # delete message from Slack channel
        try:
            response = self.client.chat_delete(
                channel=params['channel'],
                ts=params['ts']
            )
            
        except SlackApiError as e:
            raise Exception(f"Error deleting message from Slack channel '{params['channel']}' with timestamp '{params['ts']}': {e.response['error']}")

class SlackHandler(APIHandler):
    """
    A class for handling connections and interactions with Slack API.
    Agrs:
        bot_token(str): The bot token for the Slack app.
    """

    def __init__(self, name=None, **kwargs):
        """
        Initializes the connection by checking all the params are provided by the user.
        """
        super().__init__(name)

        args = kwargs.get('connection_data', {})
        self.connection_args = {}
        handler_config = Config().get('slack_handler', {})
        for k in ['token']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'SLACK_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'SLACK_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]
        self.api = None
        self.is_connected = False
        
        channels = SlackChannelsTable(self)
        self._register_table('channels', channels)

    def create_connection(self):
        """
        Creates a WebClient object to connect to the Slack API token stored in the connection_args attribute.
        """
        client = WebClient(token=self.connection_args['token'])
        return client
    
    def connect(self):
        """
        Authenticate with the Slack API using the token stored in the `token` attribute.
        """
        if self.is_connected is True:
            return self.api

        self.api = self.create_connection()
        return self.api

    def check_connection(self):
        """
        Checks the connection by calling auth_test()
        """
        response = StatusResponse(False)

        try:
            api = self.connect()

            # Call API method to check the connection
            api.auth_test()
            
            response.success = True
        except SlackApiError as e:
            response.error_message = f'Error connecting to Slack Api: {e.response["error"]}. Check token.'
            log.logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None):
        """
        Parses the query with FuncParser and calls call_slack_api and returns the result of the query as a Response object.
        """
        method_name, params = FuncParser().from_string(query_string)

        df = self.call_slack_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def call_slack_api(self, method_name: str = None, params: dict = None):
        """
        Calls specific method specified.

        Args:
            method_name: to call specific method
            params: parameters to call the method

        Returns:
            List of dictionaries as a result of the method call
        """
        api = self.connect()
        method = getattr(api, method_name)

        try:
            result = method(**params)

        except SlackApiError as e:
            error = f"Error calling method '{method_name}' with params '{params}': {e.response['error']}"
            log.logger.error(error)
            raise e

        if 'channels' in result:
            result['channels'] = self.convert_channel_data(result['channels'])

        return [result]

    def convert_channel_data(self, channels: List[dict]):
        """
        Convert the list of channel dictionaries to a format that can be easily used in the data pipeline.

        Args:
            channels: A list of channel dictionaries.

        Returns:
            A list of channel dictionaries with modified keys and values.
        """

        new_channels = []
        for channel in channels:
            new_channel = {
                'id': channel['id'],
                'name': channel['name'],
                'created': datetime.fromtimestamp(float(channel['created']))
            }
            new_channels.append(new_channel)
        return new_channels


from collections import OrderedDict
from typing import Optional
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE



import pandas as pd
import pymonetdb as mdb
from .utils.monet_get_id import *
from sqlalchemy_monetdb.dialect import MonetDialect 




class MonetDBHandler(DatabaseHandler):


    name= 'monetdb'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """ Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        
        self.kwargs = kwargs
        self.parser = parse_sql
        self.database = connection_data['database']
        self.user = connection_data['user']
        self.password = connection_data['password']
        self.schemaName = connection_data['schema_name'] if 'schema_name' in connection_data else None
        self.host = connection_data['host']
        self.port = connection_data['port']
        

        self.connection = None
        self.is_connected = False
          
    
    def connect(self):
        """ Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            Connection Object
        """
        if self.is_connected is True:
            return self.connection

        try:
            self.connection = mdb.connect(
                database=self.database, 
                hostname=self.host, 
                port=self.port, 
                username=self.user,
                password=self.password,
            )

  
            self.is_connected= True
        except Exception as e:
            log.logger.error(f"Error while connecting to {self.database}, {e}")


        return self.connection


    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected=False
        except Exception as e:
            log.logger.error(f"Error while disconnecting to {self.database}, {e}")

        return 


    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to database {self.database}, {e}!')
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode


    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        conn = self.connect()
        cur=conn.cursor() 
        try:
            cur.execute(query)
                   
            if len(cur._rows)>0 :
                result = cur.fetchall() 
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            self.connection.commit()
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.database}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
            self.connection.rollback()

        cur.close()    

        if need_to_close is True:
            self.disconnect()

        return response

    
    def query(self, query: ASTNode) -> StatusResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns: HandlerResponse
        """
        


        renderer = SqlalchemyRender(MonetDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)


    def get_tables(self) -> StatusResponse:
        """ Return list of entities
        Return list of entities that will be accesible as tables.
        Returns: HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        """
        self.connect()
        schema = schema_id(connection=self.connection,schema_name=self.schemaName)



        q = f"""
            SELECT name as TABLE_NAME
            FROM sys.tables
            WHERE system = False
            AND type = 0
            AND schema_id = {schema}
        """
 

        return self.query(q)

    
    def get_columns(self, table_name: str) -> StatusResponse:
        """ Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """
        self.connect()
        table=table_id(
            connection=self.connection,
            table_name=table_name,
            schema_name=self.schemaName
        )

        q = f"""
            SELECT  
            name as COLUMN_NAME,
            type as DATA_TYPE
            FROM sys.columns
            WHERE table_id = {table}
        """
        return self.query(q)

        





connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the MonetDB server/database.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the MonetDB server.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the MonetDB server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the MonetDB server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect MonetDB through TCP/IP'
    },
    schema_name={
        'type': ARG_TYPE.STR,
        'description': 'Specify the schema name for Listing Table '
    },

)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=50000,
    password='monetdb',
    user='monetdb',
    schemaName="sys",
    database="demo",
)
